@@ -1,16 +1,46 @@
 #ifndef __ATARI__
 #define	__ATARI__
 
-#include	"system.h"
+/*
+	=================================
+	Define Data Types on Local System
+	=================================
+*/
+
+#ifdef VMS
+#define SBYTE	char
+#define SWORD	short int
+#define SLONG	long int
+#else
+#define	SBYTE	signed char
+#define	SWORD	signed short int
+#define	SLONG	signed long int
+#endif
+
+#define	UBYTE	unsigned char
+#define	UWORD	unsigned short int
+#define	ULONG	unsigned long int
+
+typedef enum
+{
+  Atari,
+  AtariXL,
+  AtariXE,
+  Atari5200
+} Machine;
+
+extern Machine	machine;
 
 #define	ATARI_WIDTH	(384)
 #define	ATARI_HEIGHT	(192 + 24 + 24)
-#define	ATARI_TITLE	"Atari 800 Emulator, Version 0.3.0"
+#define	ATARI_TITLE	"Atari 800 Emulator, Version 0.5.0"
 
+#define NO_CART		0
 #define	NORMAL8_CART	1
 #define	NORMAL16_CART	2
-#define	OSS_SUPERCART	3
-#define	DB_SUPERCART	4
+#define AGS32_CART      3
+#define	OSS_SUPERCART	4
+#define	DB_SUPERCART	5
 
 enum ESCAPE
 {
@@ -101,67 +131,6 @@ enum ESCAPE
 	ESC_HHINIT = 0xc6
 };
 
-/*
-	==================
-	Hardware Registers
-	==================
-*/
-
-extern UBYTE	ALLPOT;
-extern UBYTE	CHACTL;
-extern UBYTE	CHBASE;
-extern UBYTE	COLBK;
-extern UBYTE	DLISTH;
-extern UBYTE	DLISTL;
-extern UBYTE	DMACTL;
-extern UBYTE	GRACTL;
-extern UBYTE	GRAFM;
-extern UBYTE	GRAFP0;
-extern UBYTE	GRAFP1;
-extern UBYTE	GRAFP2;
-extern UBYTE	GRAFP3;
-extern UBYTE	HSCROL;
-extern UBYTE	IRQEN;
-extern UBYTE	IRQST;
-extern UBYTE	KBCODE;
-extern UBYTE	M0PF;
-extern UBYTE	M0PL;
-extern UBYTE	M1PF;
-extern UBYTE	M1PL;
-extern UBYTE	M2PF;
-extern UBYTE	M2PL;
-extern UBYTE	M3PF;
-extern UBYTE	M3PL;
-extern UBYTE	NMIEN;
-extern UBYTE	NMIRES;
-extern UBYTE	NMIST;
-extern UBYTE	P0PF;
-extern UBYTE	P0PL;
-extern UBYTE	P1PF;
-extern UBYTE	P1PL;
-extern UBYTE	P2PF;
-extern UBYTE	P2PL;
-extern UBYTE	P3PF;
-extern UBYTE	P3PL;
-extern UBYTE	PACTL;
-extern UBYTE	PAL;
-extern UBYTE	PBCTL;
-extern UBYTE	PENH;
-extern UBYTE	PENV;
-extern UBYTE	PMBASE;
-extern UBYTE	PORTA;
-extern UBYTE	PORTB;
-extern UBYTE	POTGO;
-extern UBYTE	PRIOR;
-extern UBYTE	SERIN;
-extern UBYTE	SEROUT;
-extern UBYTE	SKCTL;
-extern UBYTE	SKREST;
-extern UBYTE	SKSTAT;
-extern UBYTE	STIMER;
-extern UBYTE	VDELAY;
-extern UBYTE	VSCROL;
-
 /*
 	=================
 	Joystick Position
@@ -190,6 +159,7 @@ extern UBYTE	VSCROL;
 #define AKEY_EXIT -4
 #define AKEY_BREAK -5
 #define AKEY_PIL -6
+#define AKEY_DISKCHANGE -7
 
 #define AKEY_SHFT 0x40
 #define AKEY_CTRL 0x80
@@ -379,15 +349,4 @@ extern UBYTE	VSCROL;
 #define AKEY_BACKSLASH 0x46
 #define AKEY_BAR 0x4f
 
-/*
-	==========================
-	Define Function Prototypes
-	==========================
-*/
-
-void	initialise_os ();
-void	add_esc (UWORD address, UBYTE esc_code);
-void	A800_RUN (void);
-void	ESC (UBYTE esc_code);
-
 #endif
