@@ -5,7 +5,11 @@
 #include <vga.h>
 #include <vgagl.h>
 
-static char *rcsid = "$Id: atari_svgalib.c,v 1.2 1995/12/30 22:11:06 david Exp $";
+static char *rcsid = "$Id: atari_svgalib.c,v 1.4 1996/07/19 19:53:29 david Exp $";
+
+#include "config.h"
+#include "atari.h"
+#include "colours.h"
 
 #ifdef LINUX_JOYSTICK
 #include <linux/joystick.h>
@@ -21,9 +25,6 @@ static int js1_centre_y;
 static struct JS_DATA_TYPE js_data;
 #endif
 
-#include "atari.h"
-#include "colours.h"
-
 static int	lookup[256];
 
 #define FALSE 0
@@ -33,6 +34,8 @@ static int trig0;
 static int stick0;
 static int consol;
 
+extern double deltatime;
+
 /*
    Interlace variables
 */
@@ -209,6 +212,13 @@ void Atari_DisplayScreen (UBYTE *screen)
   UBYTE *scrn_ptr;
   int ypos;
 
+/*
+ * This single command will replace all this function but doesn't allow
+ * the interlace option - maybe the interlace option should dropped.
+ *
+ * gl_putboxpart (0, 0, 320, 200, 384, 240, screen, 32, first_lno);
+ */
+
   svga_ptr = graph_mem;
   scrn_ptr = &screen[first_lno*ATARI_WIDTH+32];
 
@@ -686,6 +696,10 @@ int Atari_Keyboard (void)
 	else if (strcmp(buff, "\133\062\061\176") == 0)	/* F10 */
 	  {
 	    keycode = AKEY_NONE;
+	    if (deltatime == 0.0)
+	      deltatime = (1.0 / 8.0);
+	    else
+	      deltatime = 0.0;
 	  }
 	else if (strcmp(buff, "\133\062\063\176") == 0)	/* F11 */
 	  {
