@@ -1,6 +1,17 @@
 #include <stdio.h>
+#ifdef VMS
+#include <stat.h>
+#else
+#include <sys/stat.h>
+#endif
 
-static char *rcsid = "$Id: atari_x11.c,v 1.3 1996/01/14 22:21:46 david Exp $";
+/*
+ * Note: For SHM version check if image_data or the pixmap is needed
+ *       Check if rect, nrects, points and npoints are needed.
+ *       scanline_ptr.
+ */
+
+static char *rcsid = "$Id: atari_x11.c,v 1.23 1996/07/19 19:52:22 david Exp $";
 
 #ifdef XVIEW
 #include <xview/xview.h>
@@ -11,16 +22,49 @@ static char *rcsid = "$Id: atari_x11.c,v 1.3 1996/01/14 22:21:46 david Exp $";
 #include <xview/file_chsr.h>
 #endif
 
+#ifdef MOTIF
+#include <Xm/MainW.h>
+#include <Xm/DrawingA.h>
+#include <Xm/MessageB.h>
+#include <Xm/FileSB.h>
+
+static XtAppContext app;
+static Widget toplevel;
+static Widget main_w;
+static Widget drawing_area;
+static Widget fsel;
+#endif
+
 #include <X11/Xlib.h>
 #include <X11/Xutil.h>
 #include <X11/keysym.h>
 
+#ifndef VMS
+#include "config.h"
+#endif
+#include "atari.h"
+#include "colours.h"
+
+#ifdef FPS_MONITOR
+#include <sys/time.h>
+
+static struct timeval tp;
+static struct timezone tzp;
+
+static double basetime;
+static int totframes = 0;
+static int nframes = 0;
+#endif
+
 #ifdef SHM
 #include <sys/ipc.h>
 #include <sys/shm.h>
 #include <X11/extensions/XShm.h>
 
 static XShmSegmentInfo shminfo;
+static XImage *image;
+extern char *atari_screen;
+extern int colour_translation_table[256];
 #endif
 
 #ifdef LINUX_JOYSTICK
@@ -38,10 +82,6 @@ static int js1_centre_y;
 static struct JS_DATA_TYPE js_data;
 #endif
 
-#include "system.h"
-#include "atari.h"
-#include "colours.h"
-
 #define	FALSE	0
 #define	TRUE	1
 
@@ -62,11 +102,14 @@ static int window_height;
 static Display	*display;
 static Screen	*screen;
 static Window	window;
+#ifndef SHM
 static Pixmap	pixmap;
+#endif
 static Visual	*visual;
 
-static GC	gc;
-static GC	gc_colour[256];
+static GC gc;
+static GC gc_colour[256];
+static int colours[256];
 
 static XComposeStatus	keyboard_status;
 
@@ -74,10 +117,11 @@ static XComposeStatus	keyboard_status;
 static Frame frame;
 static Panel panel;
 static Canvas canvas;
-static Menu disk_menu;
+static Menu system_menu;
+static Menu coldstart_menu;
 static Menu consol_menu;
 static Menu options_menu;
-static Frame disk_change_chooser;
+static Frame chooser;
 
 static Frame controllers_frame;
 static Panel controllers_panel;
@@ -131,6 +175,7 @@ static int consol;
 
 extern int refresh_rate;
 extern int countdown_rate;
+extern double deltatime;
 
 int GetKeyCode (XEvent *event)
 {
@@ -144,10 +189,12 @@ int GetKeyCode (XEvent *event)
   switch (event->type)
     {
     case Expose :
+#ifndef SHM
       XCopyArea (display, pixmap, window, gc,
 		 0, 0,
 		 window_width, window_height,
 		 0, 0);
+#endif
       break;
     case KeyPress :
       switch (keysym)
@@ -198,11 +245,18 @@ int GetKeyCode (XEvent *event)
 	case XK_F7 :
 	  keycode = AKEY_BREAK;
 	  break;
+	case XK_F8 :
+	  keycode = AKEY_DISKCHANGE;
+	  break;
 	case XK_F9 :
 	  keycode = AKEY_EXIT;
 	  break;
 	case XK_F10 :
 	  keycode = AKEY_NONE;
+	  if (deltatime == 0.0)
+	    deltatime = (1.0 / 50.0);
+	  else
+	    deltatime = 0.0;
 	  break;
 	case XK_Home :
 	  keycode = 0x76;
@@ -213,6 +267,14 @@ int GetKeyCode (XEvent *event)
 	  else
 	    keycode = AKEY_INSERT_CHAR;
 	  break;
+	case XK_BackSpace :
+	  if (CONTROL)
+	    keycode = AKEY_DELETE_CHAR;
+	  else if (SHIFT)
+	    keycode = AKEY_DELETE_LINE;
+	  else
+	    keycode = AKEY_BACKSPACE;
+	  break;
 	case XK_Delete :
 	  if (CONTROL)
 	    keycode = AKEY_DELETE_CHAR;
@@ -563,7 +625,7 @@ int disk_change (char *a, char *full_filename, char *filename)
   else
     {
       if (auto_reboot)
-	xview_keycode = AKEY_COLDSTART;
+	Coldstart ();
       status = XV_OK;
     }
 
@@ -574,7 +636,10 @@ boot_callback ()
 {
   auto_reboot = TRUE;
 
-  xv_set (disk_change_chooser,
+  xv_set (chooser,
+	  FRAME_LABEL, "Disk Selector",
+	  FILE_CHOOSER_DIRECTORY, ATARI_DISK_DIR,
+	  FILE_CHOOSER_NOTIFY_FUNC, disk_change,
 	  XV_SHOW, TRUE,
 	  NULL);
 }
@@ -583,7 +648,10 @@ insert_callback ()
 {
   auto_reboot = FALSE;
 
-  xv_set (disk_change_chooser,
+  xv_set (chooser,
+	  FRAME_LABEL, "Disk Selector",
+	  FILE_CHOOSER_DIRECTORY, ATARI_DISK_DIR,
+	  FILE_CHOOSER_NOTIFY_FUNC, disk_change,
 	  XV_SHOW, TRUE,
 	  NULL);
 }
@@ -615,6 +683,105 @@ eject_callback ()
   SIO_Dismount(diskno);
 }
 
+int rom_change (char *a, char *full_filename, char *filename)
+{
+  struct stat buf;
+  int status = XV_ERROR;
+  int yesno;
+
+  stat (full_filename, &buf);
+
+  switch (buf.st_size)
+    {
+    case 0x2000 :
+      Remove_ROM ();
+      if (Insert_8K_ROM(full_filename))
+	{
+	  Coldstart ();
+	  status = XV_OK;
+	}
+      break;
+    case 0x4000 :
+      yesno = notice_prompt (panel, NULL,
+			     NOTICE_MESSAGE_STRINGS,
+			       filename,
+			       "Is this an OSS Supercartridge?",
+			       NULL,
+			     NOTICE_BUTTON_YES, "No",
+			     NOTICE_BUTTON_NO, "Yes",
+			     NULL);
+      if (yesno == NOTICE_YES)
+	{
+	  Remove_ROM ();
+	  if (Insert_16K_ROM(full_filename))
+	    {
+	      Coldstart ();
+	      status = XV_OK;
+	    }
+	}
+      else
+	{
+	  Remove_ROM ();
+	  if (Insert_OSS_ROM(full_filename))
+	    {
+	      Coldstart ();
+	      status = XV_OK;
+	    }
+	}
+      break;
+    case 0x8000 :
+      Remove_ROM ();
+      if (machine == Atari5200)
+	{
+	  if (Insert_32K_5200ROM(full_filename))
+	    {
+	      Coldstart ();
+	      status = XV_OK;
+	    }
+	}
+      else
+	{
+	  if (Insert_DB_ROM(full_filename))
+	    {
+	      Coldstart ();
+	      status = XV_OK;
+	    }
+	}
+      break;
+    default :
+      break;
+    }
+
+  return status;
+}
+
+insert_rom_callback ()
+{
+  xv_set (chooser,
+	  FRAME_LABEL, "ROM Selector",
+	  FILE_CHOOSER_DIRECTORY, ATARI_ROM_DIR,
+	  FILE_CHOOSER_NOTIFY_FUNC, rom_change,
+	  XV_SHOW, TRUE,
+	  NULL);
+}
+
+remove_rom_callback ()
+{
+  Remove_ROM ();
+  Coldstart ();
+}
+
+enable_pill_callback ()
+{
+  EnablePILL ();
+  Coldstart ();
+}
+
+exit_callback ()
+{
+  exit (1);
+}
+
 option_callback ()
 {
   consol &= 0x03;
@@ -642,12 +809,154 @@ break_callback ()
 
 reset_callback ()
 {
-  xview_keycode = AKEY_WARMSTART;
+  Warmstart ();
 }
 
 coldstart_callback ()
 {
-  xview_keycode = AKEY_COLDSTART;
+  Coldstart ();
+}
+
+coldstart_osa_callback ()
+{
+  int status;
+
+  status = Initialise_AtariOSA ();
+  if (status)
+    {
+      Menu_item menuitem;
+
+      menuitem = xv_get (consol_menu,
+			 MENU_NTH_ITEM, 4);
+
+      xv_set (menuitem,
+	      MENU_INACTIVE, TRUE,
+	      NULL);
+    }
+  else
+    {
+      notice_prompt (panel, NULL,
+		     NOTICE_MESSAGE_STRINGS,
+		       "Sorry, OS/A ROM Unavailable",
+		       NULL,
+		     NOTICE_BUTTON, "Cancel", 1,
+		     NULL);
+    }
+}
+
+coldstart_osb_callback ()
+{
+  int status;
+
+  status = Initialise_AtariOSB ();
+  if (status)
+    {
+      Menu_item menuitem;
+      
+      menuitem = xv_get (consol_menu,
+			 MENU_NTH_ITEM, 4);
+
+      xv_set (menuitem,
+	      MENU_INACTIVE, TRUE,
+	      NULL);
+    }
+  else
+    {
+      notice_prompt (panel, NULL,
+		     NOTICE_MESSAGE_STRINGS,
+		       "Sorry, OS/B ROM Unavailable",
+		       NULL,
+		     NOTICE_BUTTON, "Cancel", 1,
+		     NULL);
+    }
+}
+
+coldstart_xl_callback ()
+{
+  int status;
+
+  status = Initialise_AtariXL ();
+  if (status)
+    {
+      Menu_item menuitem;
+
+      menuitem = xv_get (consol_menu,
+			 MENU_NTH_ITEM, 4);
+      
+      xv_set (menuitem,
+	      MENU_INACTIVE, FALSE,
+	      NULL);
+    }
+  else
+    {
+      notice_prompt (panel, NULL,
+		     NOTICE_MESSAGE_STRINGS,
+		       "Sorry, XL/XE ROM Unavailable",
+		       NULL,
+		     NOTICE_BUTTON, "Cancel", 1,
+		     NULL);
+    }
+}
+
+coldstart_xe_callback ()
+{
+  int status;
+
+  status = Initialise_AtariXE ();
+  if (status)
+    {
+      Menu_item menuitem;
+
+      menuitem = xv_get (consol_menu,
+			 MENU_NTH_ITEM, 4);
+
+      xv_set (menuitem,
+	      MENU_INACTIVE, FALSE,
+	      NULL);
+    }
+  else
+    {
+      notice_prompt (panel, NULL,
+		     NOTICE_MESSAGE_STRINGS,
+		       "Sorry, XL/XE ROM Unavailable",
+		       NULL,
+		     NOTICE_BUTTON, "Cancel", 1,
+		     NULL);
+    }
+}
+
+coldstart_5200_callback ()
+{
+  int status;
+
+  status = Initialise_Atari5200 ();
+  if (status)
+    {
+      Menu_item menuitem;
+
+      menuitem = xv_get (consol_menu,
+			 MENU_NTH_ITEM, 4);
+
+      xv_set (menuitem,
+	      MENU_INACTIVE, FALSE,
+	      NULL);
+    }
+  else
+    {
+      notice_prompt (panel, NULL,
+		     NOTICE_MESSAGE_STRINGS,
+		       "Sorry, 5200 ROM Unavailable",
+		       NULL,
+		     NOTICE_BUTTON, "Cancel", 1,
+		     NULL);
+    }
+}
+
+controllers_ok_callback ()
+{
+  xv_set (controllers_frame,
+	  XV_SHOW, FALSE,
+	  NULL);
 }
 
 controllers_callback ()
@@ -659,14 +968,14 @@ controllers_callback ()
 
 void sorry_message ()
 {
-    notice_prompt (panel, NULL,
-		   NOTICE_MESSAGE_STRINGS,
-		     "Sorry, controller already assign",
-		     "to another device",
-		     NULL,
-		   NOTICE_BUTTON, "Cancel", 1,
-		   NULL);
-  }
+  notice_prompt (panel, NULL,
+		 NOTICE_MESSAGE_STRINGS,
+		   "Sorry, controller already assign",
+		   "to another device",
+		   NULL,
+		 NOTICE_BUTTON, "Cancel", 1,
+		 NULL);
+}
 
 keypad_callback ()
 {
@@ -754,6 +1063,13 @@ js1_callback ()
 }
 #endif
 
+performance_ok_callback ()
+{
+  xv_set (performance_frame,
+	  XV_SHOW, FALSE,
+	  NULL);
+}
+
 performance_callback ()
 {
   xv_set (performance_frame,
@@ -794,21 +1110,186 @@ void Atari_WhatIs (int mode)
     }
 }
 
+#ifdef MOTIF
+void motif_boot_disk (Widget fs, XtPointer client_data,
+		      XmFileSelectionBoxCallbackStruct *cbs)
+{
+  char *filename;
+
+  if (XmStringGetLtoR(cbs->value, XmSTRING_DEFAULT_CHARSET, &filename))
+    {
+      if (*filename)
+	{
+	  SIO_Dismount(1);
+	  if (SIO_Mount (1, filename))
+	    Coldstart ();
+	}
+
+      XtFree (filename);
+    }
+
+  XtUnmanageChild (fs);
+  XtPopdown (XtParent(fs));
+}
+
+void motif_fs_cancel (Widget fs)
+{
+  XtUnmanageChild (fs);
+  XtPopdown (XtParent(fs));
+}
+
+void motif_system_cback (Widget w, int item_no)
+{
+  XmString t;
+  int status;
+  char *errmsg = NULL;
+
+  switch (item_no)
+    {
+    case 0 :
+      t = XmStringCreateSimple ("/usr/local/lib/atari/DISKS/*");
+      XmFileSelectionDoSearch (fsel, t);
+      XmStringFree (t);
+      XtAddCallback (fsel, XmNokCallback, motif_boot_disk, NULL);
+      XtAddCallback (fsel, XmNcancelCallback, motif_fs_cancel, NULL);
+      XtManageChild (fsel);
+      XtPopup (XtParent(fsel), XtGrabNone);
+      break;
+    case 1 :
+    case 2 :
+    case 3 :
+      break;
+    case 4 :
+      Remove_ROM ();
+      Coldstart ();
+      break;
+    case 5 :
+      EnablePILL ();
+      Coldstart ();
+      break;
+    case 6 :
+      status = Initialise_AtariOSA ();
+      if (status == 0)
+	errmsg = "Sorry, OS/A ROM Unavailable";
+      break;
+    case 7 :
+      status = Initialise_AtariOSB ();
+      if (status == 0)
+	errmsg = "Sorry, OS/B ROM Unavailable";
+      break;
+    case 8 :
+      status = Initialise_AtariXL ();
+      if (status == 0)
+	errmsg = "Sorry, XL/XE ROM Unavailable";
+      break;
+    case 9 :
+      status = Initialise_AtariXE ();
+      if (status == 0)
+	errmsg = "Sorry, XL/XE ROM Unavailable";
+      break;
+    case 10 :
+      status = Initialise_Atari5200 ();
+      if (status == 0)
+	errmsg = "Sorry, 5200 ROM Unavailable";
+      break;
+    case 11 :
+      exit (0);
+    }
+
+  if (errmsg)
+    {
+      static Widget dialog;
+
+      if (!dialog)
+	{
+	  Arg arg[1];
+
+	  dialog = XmCreateErrorDialog (main_w, "message", arg, 0);
+
+	  XtVaSetValues (dialog,
+			 XmNdialogStyle, XmDIALOG_FULL_APPLICATION_MODAL,
+			 NULL);
+
+	  XtUnmanageChild (XmMessageBoxGetChild(dialog, XmDIALOG_OK_BUTTON));
+	  XtUnmanageChild (XmMessageBoxGetChild(dialog, XmDIALOG_HELP_BUTTON));
+	}
+
+      t = XmStringCreateSimple (errmsg);
+      XtVaSetValues (dialog,
+		     XmNmessageString, t,
+		     NULL);
+      XmStringFree (t);
+      XtManageChild (dialog);
+    }
+}
+
+void motif_consol_cback (Widget w, int item_no)
+{
+  switch (item_no)
+    {
+    case 0 :
+      consol &= 0x03; /* Option Pressed */
+      break;
+    case 1 :
+      consol &= 0x05; /* Select Pressed */
+      break;
+    case 2 :
+      consol &= 0x06; /* Start Pressed */
+      break;
+    case 3 :
+      xview_keycode = AKEY_HELP;
+      break;
+    case 4 :
+      xview_keycode = AKEY_BREAK;
+      break;
+    case 5 :
+      Warmstart ();
+      break;
+    case 6 :
+      Coldstart ();
+      break;
+    }
+}
+
+void motif_keypress (Widget w, XtPointer client_data, XEvent *event)
+{
+  int keycode;
+
+  keycode = GetKeyCode (event);
+  if (keycode != AKEY_NONE)
+    xview_keycode = keycode;
+}
+#endif
+
 void Atari_Initialise (int *argc, char *argv[])
 {
   XSetWindowAttributes	xswda;
 
   XGCValues	xgcvl;
 
-  int colours[256];
   int depth;
   int i, j;
   int mode = 0;
 
 #ifdef XVIEW
+  int ypos;
+
   xv_init (XV_INIT_ARGC_PTR_ARGV, argc, argv, NULL);
 #endif
 
+#ifdef MOTIF
+  toplevel = XtVaAppInitialize (&app, "Atari800",
+				NULL, 0,
+				argc, argv, NULL,
+				XtNtitle, ATARI_TITLE,
+				NULL);
+#endif
+
+#ifdef FPS_MONITOR
+  gettimeofday (&tp, &tzp);
+  basetime = tp.tv_sec + (tp.tv_usec / 1000000.0);
+#endif
+
   for (i=j=1;i<*argc;i++)
     {
       if (strcmp(argv[i],"-small") == 0)
@@ -842,6 +1323,14 @@ void Atari_Initialise (int *argc, char *argv[])
   NAS_Initialise (argc, argv);
 #endif
 
+#ifdef SHM
+  if (windowsize != Small)
+    {
+      printf ("X Shared memory version only supports small window\n");
+      windowsize = Small;
+    }
+#endif
+
   switch (windowsize)
     {
     case Small :
@@ -908,11 +1397,14 @@ void Atari_Initialise (int *argc, char *argv[])
   keypad_mode = mode++;
 
 #ifdef XVIEW
-  frame = (Frame)xv_create (NULL, FRAME,
+  frame = (Frame)xv_create ((Xv_opaque)NULL, FRAME,
 			    FRAME_LABEL, ATARI_TITLE,
 			    FRAME_SHOW_RESIZE_CORNER, FALSE,
 			    XV_WIDTH, window_width,
 			    XV_HEIGHT, window_height + 27,
+#ifdef FPS_MONITOR
+			    FRAME_SHOW_FOOTER, TRUE,
+#endif
 			    XV_SHOW, TRUE,
 			    NULL);
 
@@ -921,27 +1413,63 @@ void Atari_Initialise (int *argc, char *argv[])
 			    XV_SHOW, TRUE,
 			    NULL);
 
-  disk_menu = xv_create (NULL, MENU,
-			 MENU_ITEM,
-			   MENU_STRING, "Boot Disk",
-			   MENU_NOTIFY_PROC, boot_callback,
-			   NULL,
-			 MENU_ITEM,
-			   MENU_STRING, "Insert Disk",
-			   MENU_NOTIFY_PROC, insert_callback,
-			   NULL,
-			 MENU_ITEM,
-			   MENU_STRING, "Eject Disk",
-			   MENU_NOTIFY_PROC, eject_callback,
-			   NULL,
-			 NULL);
+  system_menu = xv_create ((Xv_opaque)NULL, MENU,
+			   MENU_ITEM,
+			     MENU_STRING, "Boot Disk",
+			     MENU_NOTIFY_PROC, boot_callback,
+			     NULL,
+			   MENU_ITEM,
+			     MENU_STRING, "Insert Disk",
+			     MENU_NOTIFY_PROC, insert_callback,
+			     NULL,
+			   MENU_ITEM,
+			     MENU_STRING, "Eject Disk",
+			     MENU_NOTIFY_PROC, eject_callback,
+			     NULL,
+			   MENU_ITEM,
+			     MENU_STRING, "Insert Cartridge",
+			     MENU_NOTIFY_PROC, insert_rom_callback,
+			     NULL,
+			   MENU_ITEM,
+			     MENU_STRING, "Remove Cartridge",
+			     MENU_NOTIFY_PROC, remove_rom_callback,
+			     NULL,
+			   MENU_ITEM,
+			     MENU_STRING, "Enable PILL",
+			     MENU_NOTIFY_PROC, enable_pill_callback,
+			     NULL,
+			   MENU_ITEM,
+			     MENU_STRING, "Atari 800 OS/A",
+			     MENU_NOTIFY_PROC, coldstart_osa_callback,
+			     NULL,
+			   MENU_ITEM,
+			     MENU_STRING, "Atari 800 OS/B",
+			     MENU_NOTIFY_PROC, coldstart_osb_callback,
+			     NULL,
+			   MENU_ITEM,
+			     MENU_STRING, "Atari 800XL",
+			     MENU_NOTIFY_PROC, coldstart_xl_callback,
+			     NULL,
+			   MENU_ITEM,
+			     MENU_STRING, "Atari 130XE",
+			     MENU_NOTIFY_PROC, coldstart_xe_callback,
+			     NULL,
+			   MENU_ITEM,
+			     MENU_STRING, "Atari 5200",
+			     MENU_NOTIFY_PROC, coldstart_5200_callback,
+			     NULL,
+			   MENU_ITEM,
+			     MENU_STRING, "Exit",
+			     MENU_NOTIFY_PROC, exit_callback,
+			     NULL,
+			   NULL);
 
   xv_create (panel, PANEL_BUTTON,
-	     PANEL_LABEL_STRING, "Disk",
-	     PANEL_ITEM_MENU, disk_menu,
+	     PANEL_LABEL_STRING, "System",
+	     PANEL_ITEM_MENU, system_menu,
 	     NULL);
 
-  consol_menu = (Menu)xv_create (NULL, MENU,
+  consol_menu = (Menu)xv_create ((Xv_opaque)NULL, MENU,
 				 MENU_ITEM,
 				   MENU_STRING, "Option",
 				   MENU_NOTIFY_PROC, option_callback,
@@ -978,7 +1506,7 @@ void Atari_Initialise (int *argc, char *argv[])
 	     PANEL_ITEM_MENU, consol_menu,
 	     NULL);
 
-  options_menu = (Menu)xv_create (NULL, MENU,
+  options_menu = (Menu)xv_create ((Xv_opaque)NULL, MENU,
 				  MENU_ITEM,
 				    MENU_STRING, "Controllers",
 				    MENU_NOTIFY_PROC, controllers_callback,
@@ -995,8 +1523,8 @@ void Atari_Initialise (int *argc, char *argv[])
 	     NULL);
 
   canvas = (Canvas)xv_create (frame, CANVAS,
-			      CANVAS_WIDTH, ATARI_WIDTH,
-			      CANVAS_HEIGHT, ATARI_HEIGHT,
+			      CANVAS_WIDTH, window_width,
+			      CANVAS_HEIGHT, window_height,
 			      NULL);
 /*
    =====================================
@@ -1012,7 +1540,10 @@ void Atari_Initialise (int *argc, char *argv[])
   controllers_panel = (Panel)xv_get (controllers_frame, FRAME_CMD_PANEL,
 				     NULL);
 
+  ypos = 10;
   keypad_item = (Panel_item)xv_create (controllers_panel, PANEL_CHOICE_STACK,
+				       PANEL_VALUE_X, 150,
+				       PANEL_VALUE_Y, ypos,
 				       PANEL_LAYOUT, PANEL_HORIZONTAL,
 				       PANEL_LABEL_STRING, "Numeric Keypad",
 				       PANEL_CHOICE_STRINGS,
@@ -1024,8 +1555,11 @@ void Atari_Initialise (int *argc, char *argv[])
 				       PANEL_VALUE, keypad_mode,
 				       PANEL_NOTIFY_PROC, keypad_callback,
 				       NULL);
+  ypos += 25;
 
   mouse_item = (Panel_item)xv_create (controllers_panel, PANEL_CHOICE_STACK,
+				      PANEL_VALUE_X, 150,
+				      PANEL_VALUE_Y, ypos,
 				      PANEL_LAYOUT, PANEL_HORIZONTAL,
 				      PANEL_LABEL_STRING, "Mouse",
 				      PANEL_CHOICE_STRINGS,
@@ -1046,11 +1580,14 @@ void Atari_Initialise (int *argc, char *argv[])
 				      PANEL_VALUE, mouse_mode,
 				      PANEL_NOTIFY_PROC, mouse_callback,
 				      NULL);
+  ypos += 25;
 
 #ifdef LINUX_JOYSTICK
   if (js0 != -1)
     {
       js0_item = (Panel_item)xv_create (controllers_panel, PANEL_CHOICE_STACK,
+					PANEL_VALUE_X, 150,
+					PANEL_VALUE_Y, ypos,
 					PANEL_LAYOUT, PANEL_HORIZONTAL,
 					PANEL_LABEL_STRING, "/dev/js0",
 					PANEL_CHOICE_STRINGS,
@@ -1062,11 +1599,14 @@ void Atari_Initialise (int *argc, char *argv[])
 					PANEL_VALUE, js0_mode,
 					PANEL_NOTIFY_PROC, js0_callback,
 					NULL);
+      ypos += 25;
     }
 
   if (js1 != -1)
     {
       js1_item = (Panel_item)xv_create (controllers_panel, PANEL_CHOICE_STACK,
+					PANEL_VALUE_X, 150,
+					PANEL_VALUE_Y, ypos,
 					PANEL_LAYOUT, PANEL_HORIZONTAL,
 					PANEL_LABEL_STRING, "/dev/js1",
 					PANEL_CHOICE_STRINGS,
@@ -1078,8 +1618,16 @@ void Atari_Initialise (int *argc, char *argv[])
 					PANEL_VALUE, js1_mode,
 					PANEL_NOTIFY_PROC, js1_callback,
 					NULL);
+      ypos += 25;
     }
 #endif
+
+  xv_create (controllers_panel, PANEL_BUTTON,
+	     XV_X, 130,
+	     XV_Y, 125,
+	     PANEL_LABEL_STRING, "OK",
+	     PANEL_NOTIFY_PROC, controllers_ok_callback,
+	     NULL);
 /*
    ======================================
    Create Performance Configuration Frame
@@ -1087,14 +1635,17 @@ void Atari_Initialise (int *argc, char *argv[])
 */
   performance_frame = (Frame)xv_create (frame, FRAME_CMD,
 					FRAME_LABEL, "Performance Configuration",
-					XV_WIDTH, 500,
-					XV_HEIGHT, 75,
+					XV_WIDTH, 400,
+					XV_HEIGHT, 100,
 					NULL);
 
   performance_panel = (Panel)xv_get (performance_frame, FRAME_CMD_PANEL,
 				     NULL);
 
+  ypos = 10;
   refresh_slider = (Panel_item)xv_create (performance_panel, PANEL_SLIDER,
+					  PANEL_VALUE_X, 155,
+					  PANEL_VALUE_Y, ypos,
 					  PANEL_LAYOUT, PANEL_HORIZONTAL,
 					  PANEL_LABEL_STRING, "Screen Refresh Rate",
 					  PANEL_VALUE, refresh_rate,
@@ -1104,10 +1655,13 @@ void Atari_Initialise (int *argc, char *argv[])
 					  PANEL_TICKS, 32,
 					  PANEL_NOTIFY_PROC, refresh_callback,
 					  NULL);
+  ypos += 25;
 
   countdown_slider = (Panel_item)xv_create (performance_panel, PANEL_SLIDER,
+					    PANEL_VALUE_X, 155,
+					    PANEL_VALUE_Y, ypos,
 					    PANEL_LAYOUT, PANEL_HORIZONTAL,
-					    PANEL_LABEL_STRING, "Instructions during Vertical Blank",
+					    PANEL_LABEL_STRING, "Instructions per VBI",
 					    PANEL_VALUE, countdown_rate,
 					    PANEL_MIN_VALUE, 1000,
 					    PANEL_MAX_VALUE, 10000,
@@ -1115,6 +1669,13 @@ void Atari_Initialise (int *argc, char *argv[])
 					    PANEL_TICKS, 10,
 					    PANEL_NOTIFY_PROC, countdown_callback,
 					    NULL);
+
+  xv_create (performance_panel, PANEL_BUTTON,
+	     XV_X, 180,
+	     XV_Y, 75,
+	     PANEL_LABEL_STRING, "OK",
+	     PANEL_NOTIFY_PROC, performance_ok_callback,
+	     NULL);
 /*
    ====================
    Get X Window Objects
@@ -1144,15 +1705,9 @@ void Atari_Initialise (int *argc, char *argv[])
   window = (Window)xv_get(canvas_paint_window(canvas), XV_XID);
   depth = XDefaultDepthOfScreen (screen);
 
-  disk_change_chooser = (Frame)xv_create (frame, FILE_CHOOSER,
-					  FILE_CHOOSER_DIRECTORY, ATARI_DISK_DIR,
-					  FILE_CHOOSER_TYPE, FILE_CHOOSER_OPEN,
-					  FILE_CHOOSER_NOTIFY_FUNC, disk_change,
-					  NULL);
-
-  xv_set (disk_change_chooser,
-	  FRAME_LABEL, "Disk Selector",
-	  NULL);
+  chooser = (Frame)xv_create (frame, FILE_CHOOSER,
+			      FILE_CHOOSER_TYPE, FILE_CHOOSER_OPEN,
+			      NULL);
 
   xv_set (canvas_paint_window(canvas),
 	  WIN_EVENT_PROC, event_proc,
@@ -1160,6 +1715,154 @@ void Atari_Initialise (int *argc, char *argv[])
 	  NULL);
 #endif
 
+#ifdef MOTIF
+  {
+    Widget menubar;
+
+    XmString s_system;
+    XmString s_boot_disk;
+    XmString s_insert_disk;
+    XmString s_eject_disk;
+    XmString s_insert_cart;
+    XmString s_remove_cart;
+    XmString s_enable_pill;
+    XmString s_osa;
+    XmString s_osb;
+    XmString s_osxl;
+    XmString s_osxe;
+    XmString s_os5200;
+    XmString s_exit;
+
+    XmString s_console;
+    XmString s_option;
+    XmString s_select;
+    XmString s_start;
+    XmString s_help;
+    XmString s_break;
+    XmString s_warmstart;
+    XmString s_coldstart;
+
+    main_w = XtVaCreateManagedWidget ("main_window",
+				      xmMainWindowWidgetClass, toplevel,
+				      NULL);
+
+    s_system = XmStringCreateSimple ("System");
+    s_boot_disk = XmStringCreateSimple ("Boot Disk");
+    s_insert_disk = XmStringCreateSimple ("Insert Disk");
+    s_eject_disk = XmStringCreateSimple ("Eject Disk");
+    s_insert_cart = XmStringCreateSimple ("Insert Cartridge");
+    s_remove_cart = XmStringCreateSimple ("Remove Cartridge");
+    s_enable_pill = XmStringCreateSimple ("Enable PILL");
+    s_osa = XmStringCreateSimple ("Atari 800 OS/A");
+    s_osb = XmStringCreateSimple ("Atari 800 OS/B");
+    s_osxl = XmStringCreateSimple ("Atari 800XL");
+    s_osxe = XmStringCreateSimple ("Atari 130XE");
+    s_os5200 = XmStringCreateSimple ("Atari 5200");
+    s_exit = XmStringCreateSimple ("Exit");
+
+    s_console = XmStringCreateSimple ("Console");
+    s_option = XmStringCreateSimple ("Option");
+    s_select = XmStringCreateSimple ("Select");
+    s_start = XmStringCreateSimple ("Start");
+    s_help = XmStringCreateSimple ("Help");
+    s_break = XmStringCreateSimple ("Break");
+    s_warmstart = XmStringCreateSimple ("Warmstart");
+    s_coldstart = XmStringCreateSimple ("Coldstart");
+;
+    menubar = XmVaCreateSimpleMenuBar (main_w, "menubar",
+				       XmVaCASCADEBUTTON, s_system, 'S',
+				       XmVaCASCADEBUTTON, s_console, 'C',
+				       NULL);
+
+    XmVaCreateSimplePulldownMenu (menubar, "system_menu", 0, motif_system_cback,
+				  XmVaPUSHBUTTON, s_boot_disk, 'O', NULL, NULL,
+				  XmVaPUSHBUTTON, s_insert_disk, 'I', NULL, NULL,
+				  XmVaPUSHBUTTON, s_eject_disk, 'J', NULL, NULL,
+				  XmVaSEPARATOR,
+				  XmVaPUSHBUTTON, s_insert_cart, 'N', NULL, NULL,
+				  XmVaPUSHBUTTON, s_remove_cart, 'R', NULL, NULL,
+				  XmVaPUSHBUTTON, s_enable_pill, 'P', NULL, NULL,
+				  XmVaSEPARATOR,
+				  XmVaPUSHBUTTON, s_osa, 'A', NULL, NULL,
+				  XmVaPUSHBUTTON, s_osb, 'B', NULL, NULL,
+				  XmVaPUSHBUTTON, s_osxl, 'L', NULL, NULL,
+				  XmVaPUSHBUTTON, s_osxe, 'E', NULL, NULL,
+				  XmVaPUSHBUTTON, s_os5200, '5', NULL, NULL,
+				  XmVaSEPARATOR,
+				  XmVaPUSHBUTTON, s_exit, 'x', NULL, NULL,
+				  NULL);
+
+    XmVaCreateSimplePulldownMenu (menubar, "console_menu", 1, motif_consol_cback,
+				  XmVaPUSHBUTTON, s_option, 'O', NULL, NULL,
+				  XmVaPUSHBUTTON, s_select, 't', NULL, NULL,
+				  XmVaPUSHBUTTON, s_start, 'S', NULL, NULL,
+				  XmVaSEPARATOR,
+				  XmVaPUSHBUTTON, s_help, 'H', NULL, NULL,
+				  XmVaPUSHBUTTON, s_break, 'B', NULL, NULL,
+				  XmVaSEPARATOR,
+				  XmVaPUSHBUTTON, s_warmstart, 'W', NULL, NULL,
+				  XmVaPUSHBUTTON, s_coldstart, 'C', NULL, NULL,
+				  NULL);
+
+    XmStringFree (s_system);
+    XmStringFree (s_osa);
+    XmStringFree (s_osb);
+    XmStringFree (s_osxl);
+    XmStringFree (s_osxe);
+    XmStringFree (s_os5200);
+    XmStringFree (s_exit);
+
+    XmStringFree (s_console);
+    XmStringFree (s_option);
+    XmStringFree (s_select);
+    XmStringFree (s_start);
+    XmStringFree (s_help);
+    XmStringFree (s_break);
+    XmStringFree (s_warmstart);
+    XmStringFree (s_coldstart);
+
+    XtManageChild (menubar);
+
+    fsel = XmCreateFileSelectionDialog (toplevel, "filesb", NULL, 0);
+
+    drawing_area = XtVaCreateManagedWidget ("Canvas",
+					    xmDrawingAreaWidgetClass, main_w,
+					    XmNunitType, XmPIXELS,
+					    XmNheight, window_height,
+					    XmNwidth, window_width,
+					    XmNresizePolicy, XmNONE,
+					    NULL);
+
+    XtAddEventHandler (drawing_area,
+		       KeyPressMask | KeyReleaseMask,
+		       False,
+		       motif_keypress, NULL);
+
+    XtRealizeWidget (toplevel);
+  }
+
+  display = XtDisplay (drawing_area);
+
+  window = XtWindow (drawing_area);
+
+  screen = XDefaultScreenOfDisplay (display);
+  if (!screen)
+    {
+      printf ("Unable to get screen\n");
+      exit (1);
+    }
+
+  visual = XDefaultVisualOfScreen (screen);
+  if (!visual)
+    {
+      printf ("Unable to get visual\n");
+      exit (1);
+    }
+
+  depth = XDefaultDepthOfScreen (screen);
+#endif
+
+#ifndef MOTIF
 #ifndef XVIEW
   display = XOpenDisplay (NULL);
   if (!display)
@@ -1196,6 +1899,7 @@ void Atari_Initialise (int *argc, char *argv[])
 
   XStoreName (display, window, ATARI_TITLE);
 #endif
+#endif
 
 #ifdef SHM
   {
@@ -1203,44 +1907,30 @@ void Atari_Initialise (int *argc, char *argv[])
     int minor;
     Bool pixmaps;
     Status status;
-    XImage *image;
 
     status = XShmQueryVersion (display, &major, &minor, &pixmaps);
-    if (status)
-      {
-	printf ("X Shared Memory extensions available (but not used)\n");
-	printf ("\tmajor version = %d\n", major);
-	printf ("\tminor version = %d\n", minor);
-	printf ("\tShared Memory Pixmaps %s\n",
-		(pixmaps ? "Available" : "Unavailable"));
-      }
-    else
+    if (!status)
       {
 	printf ("X Shared Memory extensions not available\n");
 	exit (1);
       }
 
-    printf ("shmseg = %x\n", shminfo.shmseg);
-    printf ("depth = %x\n", depth);
+    printf ("Using X11 Shared Memory Extensions\n");
 
     image = XShmCreateImage (display, visual, depth, ZPixmap,
 			     NULL, &shminfo, window_width, window_height);
 
-    printf ("shmseg = %x\n", shminfo.shmseg);
     shminfo.shmid = shmget (IPC_PRIVATE,
-			    window_width * window_height,
+			    (ATARI_HEIGHT+16) * ATARI_WIDTH,
 			    IPC_CREAT | 0777);
-    shminfo.shmaddr = shmat (shminfo.shmid, 0, 0);
+    shminfo.shmaddr = image->data = atari_screen = shmat (shminfo.shmid, 0, 0);
     shminfo.readOnly = False;
 
-    pixmap = XShmCreatePixmap (display, window,
-			       shminfo.shmaddr, &shminfo,
-			       window_width, window_height, depth);
-    XSync (display, False);
-    printf ("pixmap = %x\n", pixmap);
-
     XShmAttach (display, &shminfo);
+
     XSync (display, False);
+
+    shmctl (shminfo.shmid, IPC_RMID, 0);
   }
 #else
   pixmap = XCreatePixmap (display, window,
@@ -1264,6 +1954,11 @@ void Atari_Initialise (int *argc, char *argv[])
 
       colours[i] = colour.pixel;
       colours[i+1] = colour.pixel;
+
+#ifdef SHM
+      colour_translation_table[i] = colours[i];
+      colour_translation_table[i+1] = colours[i+1];
+#endif
     }
 
   for (i=0;i<256;i++)
@@ -1283,8 +1978,10 @@ void Atari_Initialise (int *argc, char *argv[])
 		  GCForeground | GCBackground,
 		  &xgcvl);
 
+#ifndef SHM
   XFillRectangle (display, pixmap, gc, 0, 0,
 		  window_width, window_height);
+#endif
 
   XMapWindow (display, window);
 
@@ -1329,16 +2026,15 @@ int Atari_Exit (int run_monitor)
 
       XSync (display, True);
 
+#ifdef SHM
+      XDestroyImage (image);
+#else
       XFreePixmap (display, pixmap);
+#endif
       XUnmapWindow (display, window);
       XDestroyWindow (display, window);
       XCloseDisplay (display);
 
-#ifdef SHM
-      shmdt (shminfo.shmaddr);
-      shmctl (shminfo.shmid, IPC_RMID, 0);
-#endif
-
 #ifdef LINUX_JOYSTICK
       if (js0 != -1)
 	close (js0);
@@ -1355,6 +2051,7 @@ int Atari_Exit (int run_monitor)
   return restart;
 }
 
+#ifndef SHM
 void Atari_ScanLine_Flush ()
 {
   if (windowsize == Small)
@@ -1380,18 +2077,19 @@ void Atari_ScanLine_Flush ()
 
   last_colour = -1;
 }
+#endif
 
 void Atari_DisplayScreen (UBYTE *screen)
 {
-  UBYTE *scanline_ptr;
-
+#ifdef SHM
+  XShmPutImage (display, window, gc, image, 0, 0, 0, 0,
+		window_width, window_height, 0);
+#else
+  UBYTE *scanline_ptr = image_data;
   int xpos;
   int ypos;
   int i;
 
-  consol = 7;
-  keypad_trig = 1;
-  scanline_ptr = image_data;
   modified = FALSE;
 
   for (ypos=0;ypos<ATARI_HEIGHT;ypos++)
@@ -1401,6 +2099,7 @@ void Atari_DisplayScreen (UBYTE *screen)
 	  UBYTE colour;
 
 	  colour = *screen++;
+
 	  if (colour != *scanline_ptr)
 	    {
 	      int flush = FALSE;
@@ -1464,12 +2163,50 @@ void Atari_DisplayScreen (UBYTE *screen)
       XCopyArea (display, pixmap, window, gc, 0, 0,
 		 window_width, window_height, 0, 0);
     }
+#endif
+
+  keypad_trig = 1;
 
 #ifdef XVIEW
+#ifdef FPS_MONITOR
+  {
+    double curtime;
+
+    gettimeofday (&tp, &tzp);
+    curtime = tp.tv_sec + (tp.tv_usec / 1000000.0);
+
+    nframes++;
+
+    if ((curtime - basetime) >= 2.0)
+      {
+	static char gash[64];
+
+	sprintf (gash, "%.2f FPS", (double)nframes / (curtime - basetime));
+
+	xv_set (frame,
+		FRAME_LEFT_FOOTER, gash,
+		NULL);
+
+	nframes = 0;
+	basetime = curtime;
+      }
+  }
+#endif
+
   notify_dispatch ();
   XFlush (display);
 #endif
 
+#ifdef MOTIF
+  while (XtAppPending(app))
+    {
+      static XEvent event;
+
+      XtAppNextEvent (app, &event);
+      XtDispatchEvent (&event);
+    }
+#endif
+
 #ifdef NAS
   NAS_UpdateSound ();
 #endif
@@ -1482,6 +2219,10 @@ int Atari_Keyboard (void)
 #ifdef XVIEW
   keycode = xview_keycode;
   xview_keycode = AKEY_NONE;
+#else
+#ifdef MOTIF
+  keycode = xview_keycode;
+  xview_keycode = AKEY_NONE;
 #else
   if (XEventsQueued (display, QueuedAfterFlush) > 0)
     {
@@ -1490,6 +2231,7 @@ int Atari_Keyboard (void)
       XNextEvent (display, &event);
       keycode = GetKeyCode (&event);
     }
+#endif
 #endif
 
   return keycode;
@@ -1554,6 +2296,7 @@ mouse_joystick (int mode)
 }
 
 #ifdef LINUX_JOYSTICK
+
 void read_joystick (int js, int centre_x, int centre_y)
 {
   const int threshold = 50;
@@ -1719,7 +2462,7 @@ int Atari_TRIG (int num)
 	trig = 1;
 
       if (js_data.buttons & 0x02)
-	xview_keycode = ' ';
+	xview_keycode = AKEY_SPACE;
     }
 
   if (num == js1_mode)
@@ -1746,13 +2489,13 @@ int Atari_POT (int num)
 
   if (num == (mouse_mode - 4))
     {
-      Window	root_return;
-      Window	child_return;
-      int	root_x_return;
-      int	root_y_return;
-      int	win_x_return;
-      int	win_y_return;
-      int	mask_return;
+      Window root_return;
+      Window child_return;
+      int root_x_return;
+      int root_y_return;
+      int win_x_return;
+      int win_y_return;
+      int mask_return;
 
       if (XQueryPointer (display, window, &root_return,
 			 &child_return, &root_x_return, &root_y_return,
@@ -1785,7 +2528,12 @@ int Atari_POT (int num)
 
 int Atari_CONSOL (void)
 {
-  return consol;
+  int temp;
+
+  temp = consol;
+  consol = 0x07;
+
+  return temp;
 }
 
 #ifndef NAS
