@@ -33,20 +33,17 @@
 
 	Call to Get Word when accessing stack - it doesn't
 	wrap around at 0x1ff correctly.
-
-	Some instruction that read and modify memory (DEC abcd etc.)
-	may not work if they operate on ROM or HARDWARE addresses.
 */
 
 #include	<stdio.h>
 #include	<stdlib.h>
 
-static char *rcsid = "$Id: cpu.c,v 1.4 1996/01/27 21:49:18 david Exp $";
+static char *rcsid = "$Id: cpu.c,v 1.13 1996/07/23 01:13:33 david Exp $";
 
 #define	FALSE	0
 #define	TRUE	1
 
-#include	"system.h"
+#include	"atari.h"
 #include	"cpu.h"
 
 /*
@@ -55,27 +52,39 @@ static char *rcsid = "$Id: cpu.c,v 1.4 1996/01/27 21:49:18 david Exp $";
 	==========================================================
 */
 
-#define UPDATE_GLOBAL_REGS s_PC=PC;s_S=S;s_A=A;s_X=X;s_Y=Y
-#define UPDATE_LOCAL_REGS PC=s_PC;S=s_S;A=s_A;X=s_X;Y=s_Y
+#define UPDATE_GLOBAL_REGS regPC=PC;regS=S;regA=A;regX=X;regY=Y
+#define UPDATE_LOCAL_REGS PC=regPC;S=regS;A=regA;X=regX;Y=regY
 
-static UWORD	s_PC;
-static UBYTE	s_A;
-static UBYTE	s_S;
-static UBYTE	s_X;
-static UBYTE	s_Y;
+UWORD regPC;
+UBYTE regA;
+UBYTE regP; /* Processor Status Byte (Partial) */
+UBYTE regS;
+UBYTE regX;
+UBYTE regY;
 
 static UBYTE	N;	/* bit7 zero (0) or bit 7 non-zero (1) */
-static UBYTE	V;
-static UBYTE	B;
-static UBYTE	D;
-static UBYTE	I;
 static UBYTE	Z;	/* zero (0) or non-zero (1) */
-static UBYTE	C;	/* zero (0) or non-zero (1) */
+static UBYTE	V;
+static UBYTE	C;	/* zero (0) or one(1) */
 
 #define RAM 0
 #define ROM 1
 #define HARDWARE 2
 
+/*
+#define PROFILE
+*/
+
+#ifdef TRACE
+extern int tron;
+#endif
+
+/*
+ * The following array is used for 6502 instruction profiling
+ */
+
+int count[256];
+
 #define MEMORY_ATTRIBUTES
 
 UBYTE memory[65536];
@@ -83,18 +92,15 @@ UBYTE memory[65536];
 static int ram_below = 0xa000;
 static int ram_above = 0xffff;
 
-int INTERRUPT;
-int ncycles;
+int IRQ;
 
 #ifdef MEMORY_ATTRIBUTES
 static UBYTE attrib[65536];
 #define	GetByte(addr)		((attrib[addr] == HARDWARE) ? Atari800_GetByte(addr) : memory[addr])
-#define	PutByte(addr,byte)	if (attrib[addr] == RAM) memory[addr] = byte; else if (attrib[addr] == HARDWARE) Atari800_PutByte(addr,byte);
-#define isRAM(addr) (attrib[addr] == RAM)
+#define	PutByte(addr,byte)	if (attrib[addr] == RAM) memory[addr] = byte; else if (attrib[addr] == HARDWARE) if (Atari800_PutByte(addr,byte)) break;
 #else
 #define GetByte(addr) ((addr < 0xd000) ? memory[addr] : ((addr > 0xd7ff) ? memory[addr] : Atari800_GetByte(addr)))
 #define PutByte(addr,byte) if ((int)addr < ram_below) memory[addr] = byte; else if ((int)addr > ram_above) memory[addr] = byte; else Atari800_PutByte(addr,byte);
-#define isRAM(addr) ((addr < ram_below) || (addr > ram_above))
 #endif
 
 #define	GetWord(addr)		((GetByte(addr+1) << 8) | GetByte(addr))
@@ -107,56 +113,50 @@ static UBYTE attrib[65536];
 	===============================================================
 */
 
-void CPU_GetStatus (CPU_Status *cpu_status)
+void CPU_GetStatus (void)
 {
-	cpu_status->PC = s_PC;
-	cpu_status->A = s_A;
-	cpu_status->S = s_S;
-	cpu_status->X = s_X;
-	cpu_status->Y = s_Y;
+  if (N)
+    SetN;
+  else
+    ClrN;
 
-	if (N)
-		cpu_status->flag.N = 1;
-	else
-		cpu_status->flag.N = 0;
+  if (Z)
+    ClrZ;
+  else
+    SetZ;
 
-	cpu_status->flag.V = V;
-	cpu_status->flag.B = B;
-	cpu_status->flag.D = D;
-	cpu_status->flag.I = I;
+  if (V)
+    SetV;
+  else
+    ClrV;
 
-	if (Z)
-		cpu_status->flag.Z = 0;
-	else
-		cpu_status->flag.Z = 1;
-
-	cpu_status->flag.C = C;
+  if (C)
+    SetC;
+  else
+    ClrC;
 }
 
-void CPU_PutStatus (CPU_Status *cpu_status)
+void CPU_PutStatus (void)
 {
-	s_PC = cpu_status->PC;
-	s_A = cpu_status->A;
-	s_S = cpu_status->S;
-	s_X = cpu_status->X;
-	s_Y = cpu_status->Y;
+  if (regP & N_FLAG)
+    N = 0x80;
+  else
+    N = 0x00;
 
-	if (cpu_status->flag.N)
-		N = 0x80;
-	else
-		N = 0x00;
+  if (regP & Z_FLAG)
+    Z = 0;
+  else
+    Z = 1;
 
-	V = cpu_status->flag.V;
-	B = cpu_status->flag.B;
-	D = cpu_status->flag.D;
-	I = cpu_status->flag.I;
+  if (regP & V_FLAG)
+    V = 1;
+  else
+    V = 0;
 
-	if (cpu_status->flag.Z)
-		Z = 0;
-	else
-		Z = 1;
-
-	C = cpu_status->flag.C;
+  if (regP & C_FLAG)
+    C = 1;
+  else
+    C = 0;
 }
 
 UBYTE BCDtoDEC[256];
@@ -170,13 +170,16 @@ void CPU_Reset (void)
 	{
 		BCDtoDEC[i] = ((i >> 4) & 0xf) * 10 + (i & 0xf);
 		DECtoBCD[i] = (((i % 100) / 10) << 4) | (i % 10);
+#ifdef PROFILE
+		count[i] = 0;
+#endif
 	}
 
-	INTERRUPT = 0x00;
+	IRQ = 0x00;
 
-	D = 0;
-	s_S = 0xff;
-	s_PC = (GetByte(0xfffd) << 8) | GetByte(0xfffc);
+	regP = 0x20; /* The unused bit is always 1 */
+	regS = 0xff;
+	regPC = (GetByte(0xfffd) << 8) | GetByte(0xfffc);
 }
 
 void SetRAM (int addr1, int addr2)
@@ -215,24 +218,43 @@ void SetHARDWARE (int addr1, int addr2)
 #endif
 }
 
-#define PHP data =  (N & 0x80) ? 0x80 : 0; \
+#define AND(t_data) data = t_data; Z = N = A &= data
+#define CMP(t_data) data = t_data; Z = N = A - data; C = (A >= data)
+#define CPX(t_data) data = t_data; Z = N = X - data; C = (X >= data);
+#define CPY(t_data) data = t_data; Z = N = Y - data; C = (Y >= data);
+#define EOR(t_data) data = t_data; Z = N = A ^= data;
+#define LDA(data) Z = N = A = data;
+#define LDX(data) Z = N = X = data;
+#define LDY(data) Z = N = Y = data;
+#define ORA(t_data) data = t_data; Z = N = A |= data
+
+#define PHP data =  (N & 0x80); \
             data |= V ? 0x40 : 0; \
-            data |= 0x20; \
-	    data |= B ? 0x10 : 0; \
-	    data |= D ? 0x08 : 0; \
-	    data |= I ? 0x04 : 0; \
+            data |= (regP & 0x3c); \
 	    data |= (Z == 0) ? 0x02 : 0; \
-	    data |= C ? 0x01 : 0; \
+	    data |= C; \
 	    memory[0x0100 + S--] = data;
 
 #define PLP data = memory[0x0100 + ++S]; \
-	    N = (data & 0x80) ? 0x80 : 0; \
+	    N = (data & 0x80); \
 	    V = (data & 0x40) ? 1 : 0; \
-	    B = (data & 0x10) ? 1 : 0; \
-	    D = (data & 0x08) ? 1 : 0; \
-	    I = (data & 0x04) ? 1 : 0; \
 	    Z = (data & 0x02) ? 0 : 1; \
-	    C = (data & 0x01) ? 1 : 0;
+	    C = (data & 0x01); \
+            regP = (data & 0x3c) | 0x20;
+
+void NMI (void)
+{
+  UBYTE S = regS;
+  UBYTE data;
+
+  memory[0x0100 + S--] = regPC >> 8;
+  memory[0x0100 + S--] = regPC & 0xff;
+  PHP;
+  SetI;
+  regPC = (memory[0xfffb] << 8) | memory[0xfffa];
+  regS = S;
+}
+
 /*
 	==============================================================
 	The first switch statement is used to determine the addressing
@@ -243,7 +265,7 @@ void SetHARDWARE (int addr1, int addr2)
 	==============================================================
 */
 
-int GO (void)
+void GO (int ncycles)
 {
 #ifdef GNU_C
   static void *opcode[256] =
@@ -336,14 +358,8 @@ int GO (void)
   UBYTE X;
   UBYTE Y;
 
-  int	cpu_status = TRUE;
-
-  UWORD	retadr;
   UWORD	addr;
   UBYTE	data;
-  SBYTE	sdata;
-
-  UBYTE	cur_opcode;
 
 /*
    This used to be in the main loop but has been removed to improve
@@ -358,31 +374,18 @@ int GO (void)
 
   UPDATE_LOCAL_REGS;
 
-  if (INTERRUPT)
+  if (IRQ)
     {
-      if (INTERRUPT & NMI_MASK)
+      if (!(regP & I_FLAG))
 	{
-	  retadr = PC;
+	  UWORD retadr = PC;
+
 	  memory[0x0100 + S--] = retadr >> 8;
 	  memory[0x0100 + S--] = retadr & 0xff;
 	  PHP;
-	  I = 1;
-	  PC = (memory[0xfffb] << 8) | memory[0xfffa];
-	  INTERRUPT &= ~NMI_MASK;
-	}
-
-      if (INTERRUPT & IRQ_MASK)
-	{
-	  if (!I)
-	    {
-	      retadr = PC;
-	      memory[0x0100 + S--] = retadr >> 8;
-	      memory[0x0100 + S--] = retadr & 0xff;
-	      PHP;
-	      I = 1;
-	      PC = (memory[0xffff] << 8) | memory[0xfffe];
-	      INTERRUPT &= ~IRQ_MASK;
-	    }
+	  SetI;
+	  PC = (memory[0xffff] << 8) | memory[0xfffe];
+	  IRQ = 0;
 	}
     }
 
@@ -431,15 +434,22 @@ int GO (void)
   while (ncycles--)
     {
 #ifdef TRACE
-      disassemble (PC, PC+1);
+      if (tron)
+	{
+	  disassemble (PC, PC+1);
+	  printf ("\tA=%2x, X=%2x, Y=%2x, S=%2x\n",
+		  A, X, Y, S);
+	}
 #endif
 
-      cur_opcode = memory[PC++];
+#ifdef PROFILE
+      count[memory[PC]]++;
+#endif
 
 #ifdef GNU_C
-      goto *opcode[cur_opcode];
+      goto *opcode[memory[PC++]];
 #else
-      switch (cur_opcode)
+      switch (memory[PC++])
 	{
 	case 0x00 :
 	  goto opcode_00;
@@ -972,31 +982,38 @@ int GO (void)
 #endif
 
     opcode_00: /* BRK */
-      retadr = PC + 1;
-      memory[0x0100 + S--] = retadr >> 8;
-      memory[0x0100 + S--] = retadr & 0xff;
-      B = 1;
-      PHP;
-      I = 1;
-      PC = (memory[0xffff] << 8) | memory[0xfffe];
+      if (!(regP & I_FLAG))
+	{
+	  UWORD retadr = PC + 1;
+	  memory[0x0100 + S--] = retadr >> 8;
+	  memory[0x0100 + S--] = retadr & 0xff;
+	  SetB;
+	  PHP;
+	  SetI;
+	  PC = (memory[0xffff] << 8) | memory[0xfffe];
+	}
       goto next;
 
     opcode_01: /* ORA (ab,x) */
       INDIRECT_X;
-      Z = N = A = A | GetByte(addr);
+      ORA (GetByte(addr));
+      goto next;
+
+    opcode_04: /* NOP (2 bytes) [unofficial] */
+      PC++;
       goto next;
 
     opcode_05: /* ORA ab */
       ZPAGE;
-      Z = N = A = A | memory[addr];
+      ORA (memory[addr]);
       goto next;
 
     opcode_06: /* ASL ab */
       ZPAGE;
       data = memory[addr];
-      C = data & 0x80;
-      Z = N = data = data << 1;
-      memory[addr] = data;
+      C = (data & 0x80) ? 1 : 0;
+      Z = N = data << 1;
+      memory[addr] = Z;
       goto next;
 
     opcode_08: /* PHP */
@@ -1004,55 +1021,60 @@ int GO (void)
       goto next;
 
     opcode_09: /* ORA #ab */
-      Z = N = A = A | memory[PC++];
+      ORA (memory[PC++]);
       goto next;
 
     opcode_0a: /* ASL */
-      C = A & 0x80;
+      C = (A & 0x80) ? 1 : 0;
       Z = N = A = A << 1;
       goto next;
 
+    opcode_0c: /* NOP (3 bytes) [unofficial] */
+      PC += 2;
+      goto next;
+
     opcode_0d: /* ORA abcd */
       ABSOLUTE;
-      Z = N = A = A | GetByte(addr);
+      ORA (GetByte(addr));
       goto next;
 
     opcode_0e: /* ASL abcd */
       ABSOLUTE;
       data = GetByte(addr);
-      C = data & 0x80;
-      Z = N = data = data << 1;
-      PutByte(addr, data);
+      C = (data & 0x80) ? 1 : 0;
+      Z = N = data << 1;
+      PutByte(addr, Z);
       goto next;
 
     opcode_10: /* BPL */
       if (!(N & 0x80))
 	{
-	  sdata = memory[PC++];
+	  SBYTE sdata = memory[PC];
 	  PC += (SWORD)sdata;
 	}
-      else
-	{
-	  PC++;
-	}
+      PC++;
       goto next;
 
     opcode_11: /* ORA (ab),y */
       INDIRECT_Y;
-      Z = N = A = A | GetByte(addr);
+      ORA (GetByte(addr));
+      goto next;
+
+    opcode_14: /* NOP (2 bytes) [unofficial] */
+      PC++;
       goto next;
 
     opcode_15: /* ORA ab,x */
       ZPAGE_X;
-      Z = N = A = A | memory[addr];
+      ORA (memory[addr]);
       goto next;
 
     opcode_16: /* ASL ab,x */
       ZPAGE_X;
       data = memory[addr];
-      C = data & 0x80;
-      Z = N = data = data << 1;
-      memory[addr] = data;
+      C = (data & 0x80) ? 1 : 0;
+      Z = N = data << 1;
+      memory[addr] = Z;
       goto next;
 
     opcode_18: /* CLC */
@@ -1061,45 +1083,53 @@ int GO (void)
 
     opcode_19: /* ORA abcd,y */
       ABSOLUTE_Y;
-      Z = N = A = A | GetByte(addr);
+      ORA (GetByte(addr));
+      goto next;
+
+    opcode_1a: /* NOP (1 byte) [unofficial] */
+      goto next;
+
+    opcode_1c: /* NOP (2 bytes) [unofficial] */
+      PC++;
       goto next;
 
     opcode_1d: /* ORA abcd,x */
       ABSOLUTE_X;
-      Z = N = A = A | GetByte(addr);
+      ORA (GetByte(addr));
       goto next;
 
     opcode_1e: /* ASL abcd,x */
       ABSOLUTE_X;
       data = GetByte(addr);
-      C = data & 0x80;
-      Z = N = data = data << 1;
-      PutByte(addr, data);
+      C = (data & 0x80) ? 1 : 0;
+      Z = N = data << 1;
+      PutByte(addr, Z);
       goto next;
 
     opcode_20: /* JSR abcd */
-      retadr = PC + 1;
-      memory[0x0100 + S--] = retadr >> 8;
-      memory[0x0100 + S--] = retadr & 0xff;
-      PC = (memory[PC+1] << 8) | memory[PC];
+      {
+	UWORD retadr = PC + 1;
+	memory[0x0100 + S--] = retadr >> 8;
+	memory[0x0100 + S--] = retadr & 0xff;
+	PC = (memory[PC+1] << 8) | memory[PC];
+      }
       goto next;
 
     opcode_21: /* AND (ab,x) */
       INDIRECT_X;
-      Z = N = A = A & GetByte(addr);
+      AND (GetByte(addr));
       goto next;
 
     opcode_24: /* BIT ab */
       ZPAGE;
-      data = memory[addr];
-      N = data;
-      V = data & 0x40;
-      Z = (A & data);
+      N = memory[addr];
+      V = N & 0x40;
+      Z = (A & N);
       goto next;
 
     opcode_25: /* AND ab */
       ZPAGE;
-      Z = N = A = A & memory[addr];
+      AND (memory[addr]);
       goto next;
 
     opcode_26: /* ROL ab */
@@ -1107,15 +1137,15 @@ int GO (void)
       data = memory[addr];
       if (C)
 	{
-	  C = data & 0x80;
-	  Z = N = data = (data << 1) | 1;
+	  C = (data & 0x80) ? 1 : 0;
+	  Z = N = (data << 1) | 1;
 	}
       else
 	{
-	  C = data & 0x80;
-	  Z = N = data = (data << 1);
+	  C = (data & 0x80) ? 1 : 0;
+	  Z = N = (data << 1);
 	}
-      memory[addr] = data;
+      memory[addr] = Z;
       goto next;
 
     opcode_28: /* PLP */
@@ -1123,33 +1153,32 @@ int GO (void)
       goto next;
 
     opcode_29: /* AND #ab */
-      Z = N = A = A & memory[PC++];
+      AND (memory[PC++]);
       goto next;
 
     opcode_2a: /* ROL */
       if (C)
 	{
-	  C = A & 0x80;
+	  C = (A & 0x80) ? 1 : 0;
 	  Z = N = A = (A << 1) | 1;
 	}
       else
 	{
-	  C = A & 0x80;
+	  C = (A & 0x80) ? 1 : 0;
 	  Z = N = A = (A << 1);
 	}
       goto next;
 
     opcode_2c: /* BIT abcd */
       ABSOLUTE;
-      data = GetByte(addr);
-      N = data;
-      V = data & 0x40;
-      Z = (A & data);
+      N = GetByte(addr);
+      V = N & 0x40;
+      Z = (A & N);
       goto next;
 
     opcode_2d: /* AND abcd */
       ABSOLUTE;
-      Z = N = A = A & GetByte(addr);
+      AND (GetByte(addr));
       goto next;
 
     opcode_2e: /* ROL abcd */
@@ -1157,37 +1186,38 @@ int GO (void)
       data = GetByte(addr);
       if (C)
 	{
-	  C = data & 0x80;
-	  Z = N = data = (data << 1) | 1;
+	  C = (data & 0x80) ? 1 : 0;
+	  Z = N = (data << 1) | 1;
 	}
       else
 	{
-	  C = data & 0x80;
-	  Z = N = data = (data << 1);
+	  C = (data & 0x80) ? 1 : 0;
+	  Z = N = (data << 1);
 	}
-      PutByte(addr, data);
+      PutByte(addr, Z);
       goto next;
 
     opcode_30: /* BMI */
       if (N & 0x80)
 	{
-	  sdata = memory[PC++];
+	  SBYTE sdata = memory[PC];
 	  PC += (SWORD)sdata;
 	}
-      else
-	{
-	  PC++;
-	}
+      PC++;
       goto next;
 
     opcode_31: /* AND (ab),y */
       INDIRECT_Y;
-      Z = N = A = A & GetByte(addr);
+      AND (GetByte(addr));
+      goto next;
+
+    opcode_34: /* NOP (2 bytes) [unofficial] */
+      PC++;
       goto next;
 
     opcode_35: /* AND ab,x */
       ZPAGE_X;
-      Z = N = A = A & memory[addr];
+      AND (memory[addr]);
       goto next;
 
     opcode_36: /* ROL ab,x */
@@ -1195,15 +1225,15 @@ int GO (void)
       data = memory[addr];
       if (C)
 	{
-	  C = data & 0x80;
-	  Z = N = data = (data << 1) | 1;
+	  C = (data & 0x80) ? 1 : 0;
+	  Z = N = (data << 1) | 1;
 	}
       else
 	{
-	  C = data & 0x80;
-	  Z = N = data = (data << 1);
+	  C = (data & 0x80) ? 1 : 0;
+	  Z = N = (data << 1);
 	}
-      memory[addr] = data;
+      memory[addr] = Z;
       goto next;
 
     opcode_38: /* SEC */
@@ -1212,12 +1242,19 @@ int GO (void)
 
     opcode_39: /* AND abcd,y */
       ABSOLUTE_Y;
-      Z = N = A = A & GetByte(addr);
+      AND (GetByte(addr));
+      goto next;
+
+    opcode_3a: /* NOP (1 byte) [unofficial] */
+      goto next;
+
+    opcode_3c: /* NOP (3 bytes) [unofficial] */
+      PC += 2;
       goto next;
 
     opcode_3d: /* AND abcd,x */
       ABSOLUTE_X;
-      Z = N = A = A & GetByte(addr);
+      AND (GetByte(addr));
       goto next;
 
     opcode_3e: /* ROL abcd,x */
@@ -1225,40 +1262,44 @@ int GO (void)
       data = GetByte(addr);
       if (C)
 	{
-	  C = data & 0x80;
-	  Z = N = data = (data << 1) | 1;
+	  C = (data & 0x80) ? 1 : 0;
+	  Z = N = (data << 1) | 1;
 	}
       else
 	{
-	  C = data & 0x80;
-	  Z = N = data = (data << 1);
+	  C = (data & 0x80) ? 1 : 0;
+	  Z = N = (data << 1);
 	}
-      PutByte(addr, data);
+      PutByte(addr, Z);
       goto next;
       
     opcode_40: /* RTI */
       PLP;
-      retadr = memory[0x0100 + ++S];
-      PC = (memory[0x0100 + ++S] << 8) | retadr;
+      data = memory[0x0100 + ++S];
+      PC = (memory[0x0100 + ++S] << 8) | data;
       goto next;
 
     opcode_41: /* EOR (ab,x) */
       INDIRECT_X;
-      Z = N = A = A ^ GetByte(addr);
+      EOR (GetByte(addr));
+      goto next;
+
+    opcode_44: /* NOP (2 bytes) [unofficial] */
+      PC++;
       goto next;
 
     opcode_45: /* EOR ab */
       ZPAGE;
-      Z = N = A = A ^ memory[addr];
+      EOR (memory[addr]);
       goto next;
 
     opcode_46: /* LSR ab */
       ZPAGE;
       data = memory[addr];
       C = data & 1;
-      Z = data = data >> 1;
+      Z = data >> 1;
       N = 0;
-      memory[addr] = data;
+      memory[addr] = Z;
       goto next;
 
     opcode_48: /* PHA */
@@ -1266,7 +1307,7 @@ int GO (void)
       goto next;
 
     opcode_49: /* EOR #ab */
-      Z = N = A = A ^ memory[PC++];
+      EOR (memory[PC++]);
       goto next;
 
     opcode_4a: /* LSR */
@@ -1282,76 +1323,83 @@ int GO (void)
 
     opcode_4d: /* EOR abcd */
       ABSOLUTE;
-      Z = N = A = A ^ GetByte(addr);
+      EOR (GetByte(addr));
       goto next;
 
     opcode_4e: /* LSR abcd */
       ABSOLUTE;
       data = GetByte(addr);
       C = data & 1;
-      Z = data = data >> 1;
+      Z = data >> 1;
       N = 0;
-      PutByte(addr, data);
+      PutByte(addr, Z);
       goto next;
 
     opcode_50: /* BVC */
       if (!V)
 	{
-	  sdata = memory[PC++];
+	  SBYTE sdata = memory[PC];
 	  PC += (SWORD)sdata;
 	}
-      else
-	{
-	  PC++;
-	}
+      PC++;
       goto next;
 
     opcode_51: /* EOR (ab),y */
       INDIRECT_Y;
-      Z = N = A = A ^ GetByte(addr);
+      EOR (GetByte(addr));
+      goto next;
+
+    opcode_54: /* NOP (2 bytes) [unofficial] */
+      PC++;
       goto next;
 
     opcode_55: /* EOR ab,x */
       ZPAGE_X;
-      Z = N = A = A ^ memory[addr];
+      EOR (memory[addr]);
       goto next;
 
     opcode_56: /* LSR ab,x */
       ZPAGE_X;
       data = memory[addr];
       C = data & 1;
-      Z = data = data >> 1;
+      Z = data >> 1;
       N = 0;
-      memory[addr] = data;
+      memory[addr] = Z;
       goto next;
 
     opcode_58: /* CLI */
-      I = 0;
+      ClrI;
       goto next;
 
     opcode_59: /* EOR abcd,y */
       ABSOLUTE_Y;
-      Z = N = A = A ^ GetByte(addr);
+      EOR (GetByte(addr));
+      goto next;
+
+    opcode_5a: /* NOP (1 byte) [unofficial] */
+      goto next;
+
+    opcode_5c: /* NOP (3 bytes) [unofficial] */
+      PC += 2;
       goto next;
 
     opcode_5d: /* EOR abcd,x */
       ABSOLUTE_X;
-      Z = N = A = A ^ GetByte(addr);
+      EOR (GetByte(addr));
       goto next;
 
     opcode_5e: /* LSR abcd,x */
       ABSOLUTE_X;
       data = GetByte(addr);
       C = data & 1;
-      Z = data = data >> 1;
+      Z = data >> 1;
       N = 0;
-      PutByte(addr, data);
+      PutByte(addr, Z);
       goto next;
 
     opcode_60: /* RTS */
-      retadr = memory[0x0100 + ++S];
-      retadr |= (memory[0x0100 + ++S] << 8);
-      PC = retadr + 1;
+      data = memory[0x0100 + ++S];
+      PC = ((memory[0x0100 + ++S] << 8) | data) + 1;
       goto next;
 
     opcode_61: /* ADC (ab,x) */
@@ -1359,6 +1407,10 @@ int GO (void)
       data = GetByte(addr);
       goto adc;
 
+    opcode_64: /* NOP (2 bytes) [unofficial] */
+      PC++;
+      goto next;
+
     opcode_65: /* ADC ab */
       ZPAGE;
       data = memory[addr];
@@ -1370,14 +1422,14 @@ int GO (void)
       if (C)
 	{
 	  C = data & 1;
-	  Z = N = data = (data >> 1) | 0x80;
+	  Z = N = (data >> 1) | 0x80;
 	}
       else
 	{
 	  C = data & 1;
-	  Z = N = data = (data >> 1);
+	  Z = N = (data >> 1);
 	}
-      memory[addr] = data;
+      memory[addr] = Z;
       goto next;
 
     opcode_68: /* PLA */
@@ -1417,26 +1469,23 @@ int GO (void)
       if (C)
 	{
 	  C = data & 1;
-	  Z = N = data = (data >> 1) | 0x80;
+	  Z = N = (data >> 1) | 0x80;
 	}
       else
 	{
 	  C = data & 1;
-	  Z = N = data = (data >> 1);
+	  Z = N = (data >> 1);
 	}
-      PutByte(addr, data);
+      PutByte(addr, Z);
       goto next;
 
     opcode_70: /* BVS */
       if (V)
 	{
-	  sdata = memory[PC++];
+	  SBYTE sdata = memory[PC];
 	  PC += (SWORD)sdata;
 	}
-      else
-	{
-	  PC++;
-	}
+      PC++;
       goto next;
 
     opcode_71: /* ADC (ab),y */
@@ -1444,6 +1493,10 @@ int GO (void)
       data = GetByte(addr);
       goto adc;
 
+    opcode_74: /* NOP (2 bytes) [unofficial] */
+      PC++;
+      goto next;
+
     opcode_75: /* ADC ab,x */
       ZPAGE_X;
       data = memory[addr];
@@ -1455,18 +1508,18 @@ int GO (void)
       if (C)
 	{
 	  C = data & 1;
-	  Z = N = data = (data >> 1) | 0x80;
+	  Z = N = (data >> 1) | 0x80;
 	}
       else
 	{
 	  C = data & 1;
-	  Z = N = data = (data >> 1);
+	  Z = N = (data >> 1);
 	}
-      memory[addr] = data;
+      memory[addr] = Z;
       goto next;
 
     opcode_78: /* SEI */
-      I = 1;
+      SetI;
       goto next;
 
     opcode_79: /* ADC abcd,y */
@@ -1474,6 +1527,13 @@ int GO (void)
       data = GetByte(addr);
       goto adc;
 
+    opcode_7a: /* NOP (1 byte) [unofficial] */
+      goto next;
+
+    opcode_7c: /* NOP (3 bytes) [unofficial] */
+      PC += 2;
+      goto next;
+
     opcode_7d: /* ADC abcd,x */
       ABSOLUTE_X;
       data = GetByte(addr);
@@ -1485,14 +1545,18 @@ int GO (void)
       if (C)
 	{
 	  C = data & 1;
-	  Z = N = data = (data >> 1) | 0x80;
+	  Z = N = (data >> 1) | 0x80;
 	}
       else
 	{
 	  C = data & 1;
-	  Z = N = data = (data >> 1);
+	  Z = N = (data >> 1);
 	}
-      PutByte(addr, data);
+      PutByte(addr, Z);
+      goto next;
+
+    opcode_80: /* NOP (2 bytes) [unofficial] */
+      PC++;
       goto next;
 
     opcode_81: /* STA (ab,x) */
@@ -1500,6 +1564,10 @@ int GO (void)
       PutByte(addr, A);
       goto next;
 
+    opcode_82: /* NOP (2 bytes) [unofficial] */
+      PC++;
+      goto next;
+
     opcode_84: /* STY ab */
       ZPAGE;
       memory[addr] = Y;
@@ -1519,6 +1587,10 @@ int GO (void)
       Z = N = --Y;
       goto next;
 
+    opcode_89: /* NOP (2 bytes) [unofficial] */
+      PC++;
+      goto next;
+
     opcode_8a: /* TXA */
       Z = N = A = X;
       goto next;
@@ -1541,13 +1613,10 @@ int GO (void)
     opcode_90: /* BCC */
       if (!C)
 	{
-	  sdata = memory[PC++];
+	  SBYTE sdata = memory[PC];
 	  PC += (SWORD)sdata;
 	}
-      else
-	{
-	  PC++;
-	}
+      PC++;
       goto next;
 
     opcode_91: /* STA (ab),y */
@@ -1589,16 +1658,16 @@ int GO (void)
       goto next;
 
     opcode_a0: /* LDY #ab */
-      Z = N = Y = memory[PC++];
+      LDY (memory[PC++]);
       goto next;
 
     opcode_a1: /* LDA (ab,x) */
       INDIRECT_X;
-      Z = N = A = GetByte(addr);
+      LDA (GetByte(addr));
       goto next;
 
     opcode_a2: /* LDX #ab */
-      Z = N = X = memory[PC++];
+      LDX (memory[PC++]);
       goto next;
 
     opcode_a3: /* LAX (ind,x) [unofficial] */
@@ -1608,17 +1677,17 @@ int GO (void)
 
     opcode_a4: /* LDY ab */
       ZPAGE;
-      Z = N = Y = memory[addr];
+      LDY (memory[addr]);
       goto next;
 
     opcode_a5: /* LDA ab */
       ZPAGE;
-      Z = N = A = memory[addr];
+      LDA (memory[addr]);
       goto next;
 
     opcode_a6: /* LDX ab */
       ZPAGE;
-      Z = N = X = memory[addr];
+      LDX (memory[addr]);
       goto next;
 
     opcode_a7: /* LAX zpage [unofficial] */
@@ -1631,7 +1700,7 @@ int GO (void)
       goto next;
 
     opcode_a9: /* LDA #ab */
-      Z = N = A = memory[PC++];
+      LDA (memory[PC++]);
       goto next;
 
     opcode_aa: /* TAX */
@@ -1640,17 +1709,17 @@ int GO (void)
 
     opcode_ac: /* LDY abcd */
       ABSOLUTE;
-      Z = N = Y = GetByte (addr);
+      LDY (GetByte (addr));
       goto next;
 
     opcode_ad: /* LDA abcd */
       ABSOLUTE;
-      Z = N = A = GetByte(addr);
+      LDA (GetByte(addr));
       goto next;
 
     opcode_ae: /* LDX abcd */
       ABSOLUTE;
-      Z = N = X = GetByte (addr);
+      LDX (GetByte (addr));
       goto next;
 
     opcode_af: /* LAX absolute [unofficial] */
@@ -1661,18 +1730,15 @@ int GO (void)
     opcode_b0: /* BCS */
       if (C)
 	{
-	  sdata = memory[PC++];
+	  SBYTE sdata = memory[PC];
 	  PC += (SWORD)sdata;
 	}
-      else
-	{
-	  PC++;
-	}
+      PC++;
       goto next;
 
     opcode_b1: /* LDA (ab),y */
       INDIRECT_Y;
-      Z = N = A = GetByte(addr);
+      LDA (GetByte(addr));
       goto next;
 
     opcode_b3: /* LAX (ind),y [unofficial] */
@@ -1682,17 +1748,17 @@ int GO (void)
 
     opcode_b4: /* LDY ab,x */
       ZPAGE_X;
-      Z = N = Y = memory[addr];
+      LDY (memory[addr]);
       goto next;
 
     opcode_b5: /* LDA ab,x */
       ZPAGE_X;
-      Z = N = A = memory[addr];
+      LDA (memory[addr]);
       goto next;
 
     opcode_b6: /* LDX ab,y */
       ZPAGE_Y;
-      Z = N = X = GetByte (addr);
+      LDX (GetByte (addr));
       goto next;
 
     opcode_b7: /* LAX zpage,y [unofficial] */
@@ -1706,7 +1772,7 @@ int GO (void)
 
     opcode_b9: /* LDA abcd,y */
       ABSOLUTE_Y;
-      Z = N = A = GetByte(addr);
+      LDA (GetByte(addr));
       goto next;
 
     opcode_ba: /* TSX */
@@ -1715,17 +1781,17 @@ int GO (void)
 
     opcode_bc: /* LDY abcd,x */
       ABSOLUTE_X;
-      Z = N = Y = GetByte (addr);
+      LDY (GetByte (addr));
       goto next;
 
     opcode_bd: /* LDA abcd,x */
       ABSOLUTE_X;
-      Z = N = A = GetByte(addr);
+      LDA (GetByte(addr));
       goto next;
 
     opcode_be: /* LDX abcd,y */
       ABSOLUTE_Y;
-      Z = N = X = GetByte (addr);
+      LDX (GetByte (addr));
       goto next;
 
     opcode_bf: /* LAX absolute,y [unofficial] */
@@ -1734,30 +1800,26 @@ int GO (void)
       goto next;
 
     opcode_c0: /* CPY #ab */
-      data = memory[PC++];
-      Z = N = Y - data;
-      C = (Y >= data);
+      CPY (memory[PC++]);
       goto next;
 
     opcode_c1: /* CMP (ab,x) */
       INDIRECT_X;
-      data = GetByte(addr);
-      Z = N = A - data;
-      C = (A >= data);
+      CMP (GetByte(addr));
+      goto next;
+
+    opcode_c2: /* NOP (2 bytes) [unofficial] */
+      PC++;
       goto next;
 
     opcode_c4: /* CPY ab */
       ZPAGE;
-      data = memory[addr];
-      Z = N = Y - data;
-      C = (Y >= data);
+      CPY (memory[addr]);
       goto next;
 
     opcode_c5: /* CMP ab */
       ZPAGE;
-      data = memory[addr];
-      Z = N = A - data;
-      C = (A >= data);
+      CMP (memory[addr]);
       goto next;
 
     opcode_c6: /* DEC ab */
@@ -1770,9 +1832,7 @@ int GO (void)
       goto next;
 
     opcode_c9: /* CMP #ab */
-      data = memory[PC++];
-      Z = N = A - data;
-      C = (A >= data);
+      CMP (memory[PC++]);
       goto next;
 
     opcode_ca: /* DEX */
@@ -1781,46 +1841,42 @@ int GO (void)
 
     opcode_cc: /* CPY abcd */
       ABSOLUTE;
-      data = GetByte(addr);
-      Z = N = Y - data;
-      C = (Y >= data);
+      CPY (GetByte(addr));
       goto next;
 
     opcode_cd: /* CMP abcd */
       ABSOLUTE;
-      data = GetByte(addr);
-      Z = N = A - data;
-      C = (A >= data);
+      CMP (GetByte(addr));
       goto next;
 
     opcode_ce: /* DEC abcd */
       ABSOLUTE;
-      if (isRAM(addr))
-	Z = N = --memory[addr];
+      Z = N = GetByte(addr) - 1;
+      PutByte(addr, Z);
       goto next;
 
     opcode_d0: /* BNE */
-      if (Z != 0)
+      if (Z)
 	{
-	  sdata = memory[PC++];
+	  SBYTE sdata = memory[PC];
+
 	  PC += (SWORD)sdata;
 	}
-      else
-	{
-	  PC++;
-	}
+      PC++;
       goto next;
 
     opcode_d1: /* CMP (ab),y */
       INDIRECT_Y;
-      data = GetByte(addr);
-      Z = N = A - data;
-      C = (A >= data);
+      CMP (GetByte(addr));
+      goto next;
+
+    opcode_d4: /* NOP (2 bytes) [unofficial] */
+      PC++;
       goto next;
 
     opcode_d5: /* CMP ab,x */
       ZPAGE_X;
-      data = memory[addr];
+      CMP (memory[addr]);
       Z = N = A - data;
       C = (A >= data);
       goto next;
@@ -1831,33 +1887,34 @@ int GO (void)
       goto next;
 
     opcode_d8: /* CLD */
-      D = 0;
+      ClrD;
       goto next;
 
     opcode_d9: /* CMP abcd,y */
       ABSOLUTE_Y;
-      data = GetByte(addr);
-      Z = N = A - data;
-      C = (A >= data);
+      CMP (GetByte(addr));
+      goto next;
+
+    opcode_da: /* NOP (1 byte) [unofficial] */
+      goto next;
+
+    opcode_dc: /* NOP (3 bytes) [unofficial] */
+      PC += 2;
       goto next;
 
     opcode_dd: /* CMP abcd,x */
       ABSOLUTE_X;
-      data = GetByte(addr);
-      Z = N = A - data;
-      C = (A >= data);
+      CMP (GetByte(addr));
       goto next;
 
     opcode_de: /* DEC abcd,x */
       ABSOLUTE_X;
-      if (isRAM(addr))
-	Z = N = --memory[addr];
+      Z = N = GetByte(addr) - 1;
+      PutByte(addr, Z);
       goto next;
 
     opcode_e0: /* CPX #ab */
-      data = memory[PC++];
-      Z = N = X - data;
-      C = (X >= data);
+      CPX (memory[PC++]);
       goto next;
 
     opcode_e1: /* SBC (ab,x) */
@@ -1865,11 +1922,13 @@ int GO (void)
       data = GetByte(addr);
       goto sbc;
 
+    opcode_e2: /* NOP (2 bytes) [unofficial] */
+      PC++;
+      goto next;
+
     opcode_e4: /* CPX ab */
       ZPAGE;
-      data = memory[addr];
-      Z = N = X - data;
-      C = (X >= data);
+      CPX (memory[addr]);
       goto next;
 
     opcode_e5: /* SBC ab */
@@ -1895,9 +1954,7 @@ int GO (void)
 
     opcode_ec: /* CPX abcd */
       ABSOLUTE;
-      data = GetByte (addr);
-      Z = N = X - data;
-      C = (X >= data);
+      CPX (GetByte (addr));
       goto next;
 
     opcode_ed: /* SBC abcd */
@@ -1907,20 +1964,17 @@ int GO (void)
 
     opcode_ee: /* INC abcd */
       ABSOLUTE;
-      if (isRAM(addr))
-	Z = N = ++memory[addr];
+      Z = N = GetByte(addr) + 1;
+      PutByte(addr, Z);
       goto next;
 
     opcode_f0: /* BEQ */
-      if (Z == 0)
+      if (!Z)
 	{
-	  sdata = memory[PC++];
+	  SBYTE sdata = memory[PC];
 	  PC += (SWORD)sdata;
 	}
-      else
-	{
-	  PC++;
-	}
+      PC++;
       goto next;
 
     opcode_f1: /* SBC (ab),y */
@@ -1928,6 +1982,10 @@ int GO (void)
       data = GetByte(addr);
       goto sbc;
 
+    opcode_f4: /* NOP (2 bytes) [unofficial] */
+      PC++;
+      goto next;
+
     opcode_f5: /* SBC ab,x */
       ZPAGE_X;
       data = memory[addr];
@@ -1939,7 +1997,7 @@ int GO (void)
       goto next;
 
     opcode_f8: /* SED */
-      D = 1;
+      SetD;
       goto next;
 
     opcode_f9: /* SBC abcd,y */
@@ -1947,6 +2005,13 @@ int GO (void)
       data = GetByte(addr);
       goto sbc;
 
+    opcode_fa: /* NOP (1 byte) [unofficial] */
+      goto next;
+
+    opcode_fc: /* NOP (3 bytes) [unofficial] */
+      PC += 2;
+      goto next;
+
     opcode_fd: /* SBC abcd,x */
       ABSOLUTE_X;
       data = GetByte(addr);
@@ -1954,42 +2019,40 @@ int GO (void)
 
     opcode_fe: /* INC abcd,x */
       ABSOLUTE_X;
-      if (isRAM(addr))
-	Z = N = ++memory[addr];
+      Z = N = GetByte(addr) + 1;
+      PutByte(addr, Z);
       goto next;
 
     opcode_d2: /* ESCRTS #ab (JAM) */
       data = memory[PC++];
       UPDATE_GLOBAL_REGS;
+      CPU_GetStatus();
       Escape (data);
+      CPU_PutStatus();
       UPDATE_LOCAL_REGS;
-      retadr = memory[0x0100 + ++S];
-      retadr |= (memory[0x0100 + ++S] << 8);
-      PC = retadr + 1;
+      data = memory[0x0100 + ++S];
+      PC = ((memory[0x0100 + ++S] << 8) | data) + 1;
       goto next;
 
     opcode_f2: /* ESC #ab (JAM) */
     opcode_ff: /* ESC #ab */
       data = memory[PC++];
       UPDATE_GLOBAL_REGS;
+      CPU_GetStatus();
       Escape (data);
+      CPU_PutStatus();
       UPDATE_LOCAL_REGS;
       goto next;
 
     opcode_02:
     opcode_03:
-    opcode_04:
     opcode_07:
     opcode_0b:
-    opcode_0c:
     opcode_0f:
     opcode_12:
     opcode_13:
-    opcode_14:
     opcode_17:
-    opcode_1a:
     opcode_1b:
-    opcode_1c:
     opcode_1f:
     opcode_22:
     opcode_23:
@@ -1998,45 +2061,31 @@ int GO (void)
     opcode_2f:
     opcode_32:
     opcode_33:
-    opcode_34:
     opcode_37:
-    opcode_3a:
     opcode_3b:
-    opcode_3c:
     opcode_3f:
     opcode_42:
     opcode_43:
-    opcode_44:
     opcode_47:
     opcode_4b:
     opcode_4f:
     opcode_52:
     opcode_53:
-    opcode_54:
     opcode_57:
-    opcode_5a:
     opcode_5b:
-    opcode_5c:
     opcode_5f:
     opcode_62:
     opcode_63:
-    opcode_64:
     opcode_67:
     opcode_6b:
     opcode_6f:
     opcode_72:
     opcode_73:
-    opcode_74:
     opcode_77:
-    opcode_7a:
     opcode_7b:
-    opcode_7c:
     opcode_7f:
-    opcode_80:
-    opcode_82:
     opcode_83:
     opcode_87:
-    opcode_89:
     opcode_8b:
     opcode_8f:
     opcode_92:
@@ -2049,93 +2098,71 @@ int GO (void)
     opcode_ab:
     opcode_b2:
     opcode_bb:
-    opcode_c2:
     opcode_c3:
     opcode_c7:
     opcode_cb:
     opcode_cf:
     opcode_d3:
-    opcode_d4:
     opcode_d7:
-    opcode_da:
     opcode_db:
-    opcode_dc:
     opcode_df:
-    opcode_e2:
     opcode_e3:
     opcode_e7:
     opcode_eb:
     opcode_ef:
     opcode_f3:
-    opcode_f4:
     opcode_f7:
-    opcode_fa:
     opcode_fb:
-    opcode_fc:
       UPDATE_GLOBAL_REGS;
       fprintf (stderr,"*** Invalid Opcode %02x at address %04x\n",
-	       cur_opcode, PC-1);
-      cpu_status = FALSE;
+	       memory[PC-1], PC-1);
       ncycles = 1;
       break;
 
     adc:
-      if (C) C = 1;
-      if (!D)
+      if (!(regP & D_FLAG))
 	{
-	  UBYTE	old_A;
 	  UWORD	temp;
-	  
-	  old_A = A;
 
-	  temp = (UWORD)A + (UWORD)data + (UWORD)C;
-
-	  Z = N = A = temp & 0xff;
+	  Z = N = temp = (UWORD)A + (UWORD)data + (UWORD)C;
 
 	  C = temp >> 8;
-	  V = (old_A ^ A) & 0x80;
+	  V = (Z ^ A) & 0x80;
+	  A = Z;
 	}
       else
 	{
-	  UBYTE	old_A;
 	  int	bcd1, bcd2;
 
-	  old_A = A;
-
 	  bcd1 = BCDtoDEC[A];
 	  bcd2 = BCDtoDEC[data];
 
 	  bcd1 += bcd2 + C;
 
-	  Z = N = A = DECtoBCD[bcd1];
+	  Z = N = DECtoBCD[bcd1];
 
 	  C = (bcd1 > 99);
-	  V = (old_A ^ A) & 0x80;
+	  V = (Z ^ A) & 0x80;
+	  A = Z;
 	}
       goto next;
 
     sbc:
-      if (C) C = 1;
-      if (!D)
+      if (!(regP & D_FLAG))
 	{
 	  UWORD	temp;
-	  UBYTE	old_A;
-
-	  old_A = A;
 
 	  temp = (UWORD)A - (UWORD)data - (UWORD)!C;
 
-	  Z = N = A = temp & 0xff;
+	  Z = N = temp & 0xff;
 
-	  C = (old_A < ((UWORD)data + (UWORD)!C)) ? 0 : 1;
-	  V = (old_A ^ A) & 0x80;
+	  C = (A < ((UWORD)data + (UWORD)!C)) ? 0 : 1;
+	  V = (Z ^ A) & 0x80;
+	  A = Z;
 	}
       else
 	{
 	  int	bcd1, bcd2;
-	  UBYTE	old_A;
-
-	  old_A = A;
 
 	  bcd1 = BCDtoDEC[A];
 	  bcd2 = BCDtoDEC[data];
@@ -2143,10 +2170,11 @@ int GO (void)
 	  bcd1 = bcd1 - bcd2 - !C;
 
 	  if (bcd1 < 0) bcd1 = 100 - (-bcd1);
-	  Z = N = A = DECtoBCD[bcd1];
+	  Z = N = DECtoBCD[bcd1];
 
-	  C = (old_A < (data + !C)) ? 0 : 1;
-	  V = (old_A ^ A) & 0x80;
+	  C = (A < (data + !C)) ? 0 : 1;
+	  V = (Z ^ A) & 0x80;
+	  A = Z;
 	}
       goto next;
 
@@ -2155,6 +2183,4 @@ int GO (void)
     }
 
   UPDATE_GLOBAL_REGS;
-
-  return cpu_status;
 }
