@@ -1,12 +1,17 @@
 #include	<stdio.h>
 #include	<stdlib.h>
 #include	<string.h>
-#include	<fcntl.h>
 #include	<ctype.h>
 
+#ifdef VMS
+#include	<unixio.h>
+#include	<file.h>
+#else
+#include	<fcntl.h>
 #ifndef	AMIGA
 #include	<unistd.h>
 #endif
+#endif
 
 #define	DO_DIR
 
@@ -22,15 +27,20 @@
 #include	<dirent.h>
 #endif
 
-static char *rcsid = "$Id: devices.c,v 1.8 1996/02/05 18:16:22 david Exp $";
+static char *rcsid = "$Id: devices.c,v 1.13 1996/07/18 00:33:05 david Exp $";
 
 #define FALSE   0
 #define TRUE    1
 
-#include "system.h"
-#ifndef AMIGA
+#ifdef AMIGA
+#define PRINT_COMMAND "COPY %s PAR:"
+#else
+#ifdef VMS
+#define PRINT_COMMAND "$ PRINT/DELETE %s"
+#else
 #include "config.h"
 #endif
+#endif
 #include "atari.h"
 #include "cpu.h"
 
@@ -68,8 +78,6 @@ static char	filename[64];
 static DIR	*dp = NULL;
 #endif
 
-static CPU_Status	cpu_status;
-
 void Device_Initialise (int *argc, char *argv[])
 {
   int i;
@@ -98,6 +106,28 @@ void Device_Initialise (int *argc, char *argv[])
   *argc = j;
 }
 
+int Device_isvalid (char ch)
+{
+  int valid;
+
+  if (isalnum(ch))
+    valid = TRUE;
+  else
+    switch (ch)
+      {
+      case ':' :
+      case '.' :
+      case '_' :
+	valid = TRUE;
+	break;
+      default :
+	valid = FALSE;
+	break;
+      }
+
+  return valid;
+}
+
 Device_GetFilename ()
 {
   int bufadr;
@@ -106,7 +136,7 @@ Device_GetFilename ()
 
   bufadr = (memory[ICBAHZ] << 8) | memory[ICBALZ];
 
-  while (isprint(memory[bufadr]))
+  while (Device_isvalid(memory[bufadr]))
     {
       int byte = memory[bufadr];
 
@@ -166,8 +196,6 @@ void Device_HHOPEN (void)
   if (devbug)
     printf ("HHOPEN\n");
 
-  CPU_GetStatus (&cpu_status);
-
   fid = memory[0x2e] >> 4;
 
   if (fp[fid])
@@ -202,10 +230,14 @@ void Device_HHOPEN (void)
       flag[fid] = FALSE;
     }
 
+#ifdef VMS
+/* Assumes H[devnum] is a directory _logical_, not an explicit directory
+   specification! */
+  sprintf (fname, "%s:%s", H[devnum], filename);
+#else
   sprintf (fname, "%s/%s", H[devnum], filename);
-/*
-  printf ("devnum = %d, filename = %s\n", devnum, fname);
-*/
+#endif
+
   temp = memory[ICAX1Z];
 
   switch (temp)
@@ -214,13 +246,13 @@ void Device_HHOPEN (void)
       fp[fid] = fopen (fname, "r");
       if (fp[fid])
 	{
-	  cpu_status.Y = 1;
-	  cpu_status.flag.N = 0;
+	  regY = 1;
+	  ClrN;
 	}
       else
 	{
-	  cpu_status.Y = 170;
-	  cpu_status.flag.N = 1;
+	  regY = 170;
+	  SetN;
 	}
       break;
     case 6 :
@@ -241,8 +273,8 @@ void Device_HHOPEN (void)
 	      
 	      closedir (dp);
 
-	      cpu_status.Y = 1;
-	      cpu_status.flag.N = 0;
+	      regY = 1;
+	      ClrN;
 
 	      rewind (fp[fid]);
 
@@ -250,8 +282,8 @@ void Device_HHOPEN (void)
 	    }
 	  else
 	    {
-	      cpu_status.Y = 163;
-	      cpu_status.flag.N = 1;
+	      regY = 163;
+	      SetN;
 	      fclose (fp[fid]);
 	      fp[fid] = NULL;
 	    }
@@ -259,30 +291,28 @@ void Device_HHOPEN (void)
       else
 #endif
 	{
-	  cpu_status.Y = 163;
-	  cpu_status.flag.N = 1;
+	  regY = 163;
+	  SetN;
 	}
       break;
     case 8 :
       fp[fid] = fopen (fname, "w");
       if (fp[fid])
 	{
-	  cpu_status.Y = 1;
-	  cpu_status.flag.N = 0;
+	  regY = 1;
+	  ClrN;
 	}
       else
 	{
-	  cpu_status.Y = 170;
-	  cpu_status.flag.N = 1;
+	  regY = 170;
+	  SetN;
 	}
       break;
     default :
-      cpu_status.Y = 163;
-      cpu_status.flag.N = 1;
+      regY = 163;
+      SetN;
       break;
     }
-  
-  CPU_PutStatus (&cpu_status);
 }
 
 void Device_HHCLOS (void)
@@ -290,8 +320,6 @@ void Device_HHCLOS (void)
   if (devbug)
     printf ("HHCLOS\n");
 
-  CPU_GetStatus (&cpu_status);
-
   fid = memory[0x2e] >> 4;
 
   if (fp[fid])
@@ -300,10 +328,8 @@ void Device_HHCLOS (void)
       fp[fid] = NULL;
     }
 
-  cpu_status.Y = 1;
-  cpu_status.flag.N = 0;
-
-  CPU_PutStatus (&cpu_status);
+  regY = 1;
+  ClrN;
 }
 
 void Device_HHREAD (void)
@@ -311,8 +337,6 @@ void Device_HHREAD (void)
   if (devbug)
     printf ("HHREAD\n");
 
-  CPU_GetStatus (&cpu_status);
-
   fid = memory[0x2e] >> 4;
 
   if (fp[fid])
@@ -334,23 +358,21 @@ void Device_HHREAD (void)
 		}
 	    }
 
-	  cpu_status.A = ch;
-	  cpu_status.Y = 1;
-	  cpu_status.flag.N = 0;
+	  regA = ch;
+	  regY = 1;
+	  ClrN;
 	}
       else
 	{
-	  cpu_status.Y = 136;
-	  cpu_status.flag.N = 1;
+	  regY = 136;
+	  SetN;
 	}
     }
   else
     {
-      cpu_status.Y = 163;
-      cpu_status.flag.N = 1;
+      regY = 163;
+      SetN;
     }
-
-  CPU_PutStatus (&cpu_status);
 }
 
 void Device_HHWRIT (void)
@@ -358,15 +380,13 @@ void Device_HHWRIT (void)
   if (devbug)
     printf ("HHWRIT\n");
 
-  CPU_GetStatus (&cpu_status);
-
   fid = memory[0x2e] >> 4;
 
   if (fp[fid])
     {
       int ch;
 
-      ch = cpu_status.A;
+      ch = regA;
       if (flag[fid])
 	{
 	  switch (ch)
@@ -379,16 +399,14 @@ void Device_HHWRIT (void)
 	    }
 	}
       fputc (ch, fp[fid]);
-      cpu_status.Y = 1;
-      cpu_status.flag.N = 0;
+      regY = 1;
+      ClrN;
     }
   else
     {
-      cpu_status.Y = 163;
-      cpu_status.flag.N = 1;
+      regY = 163;
+      SetN;
     }
-
-  CPU_PutStatus (&cpu_status);
 }
 
 void Device_HHSTAT (void)
@@ -396,14 +414,10 @@ void Device_HHSTAT (void)
   if (devbug)
     printf ("HHSTAT\n");
 
-  CPU_GetStatus (&cpu_status);
-
   fid = memory[0x2e] >> 4;
 
-  cpu_status.Y = 146;
-  cpu_status.flag.N = 1;
-  
-  CPU_PutStatus (&cpu_status);
+  regY = 146;
+  SetN;
 }
 
 void Device_HHSPEC (void)
@@ -411,8 +425,6 @@ void Device_HHSPEC (void)
   if (devbug)
     printf ("HHSPEC\n");
 
-  CPU_GetStatus (&cpu_status);
-
   fid = memory[0x2e] >> 4;
 
   switch (memory[ICCOMZ])
@@ -443,10 +455,8 @@ void Device_HHSPEC (void)
       break;
     }
 
-  cpu_status.Y = 146;
-  cpu_status.flag.N = 1;
-
-  CPU_PutStatus (&cpu_status);
+  regY = 146;
+  SetN;
 }
 
 void Device_HHINIT (void)
@@ -464,8 +474,6 @@ void Device_PHOPEN (void)
   if (devbug)
     printf ("PHOPEN\n");
 
-  CPU_GetStatus (&cpu_status);
-
   if (phfd != -1)
     Device_PHCLOS ();
 
@@ -473,16 +481,14 @@ void Device_PHOPEN (void)
   phfd = open (spool_file, O_CREAT | O_TRUNC | O_WRONLY, 0777);
   if (phfd != -1)
     {
-	cpu_status.Y = 1;
-	cpu_status.flag.N = 0;
+	regY = 1;
+	ClrN;
     }
   else
     {
-	cpu_status.Y = 130;
-	cpu_status.flag.N = 1;
+	regY = 130;
+	SetN;
     }
-
-  CPU_PutStatus (&cpu_status);
 }
 
 void Device_PHCLOS (void)
@@ -490,8 +496,6 @@ void Device_PHCLOS (void)
   if (devbug)
     printf ("PHCLOS\n");
 
-  CPU_GetStatus (&cpu_status);
-
   if (phfd != -1)
     {
       char command[256];
@@ -502,20 +506,20 @@ void Device_PHCLOS (void)
       sprintf (command, PRINT_COMMAND, spool_file);
       system (command);
 
+#ifndef VMS
       status = unlink (spool_file);
       if (status == -1)
 	{
 	  perror (spool_file);
 	  exit (1);
 	}
+#endif
 
       phfd = -1;
     }
 
-  cpu_status.Y = 1;
-  cpu_status.flag.N = 0;
-
-  CPU_PutStatus (&cpu_status);
+  regY = 1;
+  ClrN;
 }
 
 void Device_PHREAD (void)
@@ -523,12 +527,8 @@ void Device_PHREAD (void)
   if (devbug)
     printf ("PHREAD\n");
 
-  CPU_GetStatus (&cpu_status);
-
-  cpu_status.Y = 146;
-  cpu_status.flag.N = 1;
-
-  CPU_PutStatus (&cpu_status);
+  regY = 146;
+  SetN;
 }
 
 void Device_PHWRIT (void)
@@ -539,34 +539,27 @@ void Device_PHWRIT (void)
   if (devbug)
     printf ("PHWRIT\n");
 
-  CPU_GetStatus (&cpu_status);
-
-  byte = cpu_status.A;
+  byte = regA;
   if (byte == 0x9b)
     byte = '\n';
 
   status = write (phfd, &byte, 1);
   if (status == 1)
     {
-      cpu_status.Y = 1;
-      cpu_status.flag.N = 0;
+      regY = 1;
+      ClrN;
     }
   else
     {
-      cpu_status.Y = 144;
-      cpu_status.flag.N = 1;
+      regY = 144;
+      SetN;
     }
-
-  CPU_PutStatus (&cpu_status);
 }
 
 void Device_PHSTAT (void)
 {
   if (devbug)
     printf ("PHSTAT\n");
-
-  CPU_GetStatus (&cpu_status);
-  CPU_PutStatus (&cpu_status);
 }
 
 void Device_PHSPEC (void)
@@ -574,12 +567,8 @@ void Device_PHSPEC (void)
   if (devbug)
     printf ("PHSPEC\n");
 
-  CPU_GetStatus (&cpu_status);
-
-  cpu_status.Y = 1;
-  cpu_status.flag.N = 0;
-
-  CPU_PutStatus (&cpu_status);
+  regY = 1;
+  ClrN;
 }
 
 void Device_PHINIT (void)
@@ -587,13 +576,9 @@ void Device_PHINIT (void)
   if (devbug)
     printf ("PHINIT\n");
 
-  CPU_GetStatus (&cpu_status);
-
   phfd = -1;
-  cpu_status.Y = 1;
-  cpu_status.flag.N = 0;
-
-  CPU_PutStatus (&cpu_status);
+  regY = 1;
+  ClrN;
 }
 
 void Device_KHOPEN (void)
@@ -755,11 +740,8 @@ void Device_SHOPEN (void)
   if (devbug)
     printf ("SHOPEN\n");
 
-  CPU_GetStatus (&cpu_status);
-
-  cpu_status.Y = 1;
-  cpu_status.flag.N = 0;
-  CPU_PutStatus (&cpu_status);
+  regY = 1;
+  ClrN;
 }
 
 void Device_SHCLOS (void)
@@ -767,10 +749,8 @@ void Device_SHCLOS (void)
   if (devbug)
     printf ("SHCLOS\n");
 
-  CPU_GetStatus (&cpu_status);
-  cpu_status.Y = 1;
-  cpu_status.flag.N = 0;
-  CPU_PutStatus (&cpu_status);
+  regY = 1;
+  ClrN;
 }
 
 void Device_SHREAD (void)
@@ -778,10 +758,8 @@ void Device_SHREAD (void)
   if (devbug)
     printf ("SHREAD\n");
 
-  CPU_GetStatus (&cpu_status);
-  cpu_status.Y = 1;
-  cpu_status.flag.N = 0;
-  CPU_PutStatus (&cpu_status);
+  regY = 1;
+  ClrN;
 }
 
 void Device_SHWRIT (void)
@@ -789,11 +767,9 @@ void Device_SHWRIT (void)
   if (devbug)
     printf ("SHWRIT\n");
 
-  CPU_GetStatus (&cpu_status);
-  Device_PutChar (cpu_status.A);
-  cpu_status.Y = 1;
-  cpu_status.flag.N = 0;
-  CPU_PutStatus (&cpu_status);
+  Device_PutChar (regA);
+  regY = 1;
+  ClrN;
 }
 
 void Device_SHSTAT (void)
@@ -801,10 +777,8 @@ void Device_SHSTAT (void)
   if (devbug)
     printf ("SHSTAT\n");
 
-  CPU_GetStatus (&cpu_status);
-  cpu_status.Y = 1;
-  cpu_status.flag.N = 0;
-  CPU_PutStatus (&cpu_status);
+  regY = 1;
+  ClrN;
 }
 
 void Device_SHSPEC (void)
@@ -812,10 +786,8 @@ void Device_SHSPEC (void)
   if (devbug)
     printf ("SHSPEC\n");
 
-  CPU_GetStatus (&cpu_status);
-  cpu_status.Y = 1;
-  cpu_status.flag.N = 0;
-  CPU_PutStatus (&cpu_status);
+  regY = 1;
+  ClrN;
 }
 
 void Device_SHINIT (void)
@@ -823,10 +795,8 @@ void Device_SHINIT (void)
   if (devbug)
     printf ("SHINIT\n");
 
-  CPU_GetStatus (&cpu_status);
-  cpu_status.Y = 1;
-  cpu_status.flag.N = 0;
-  CPU_PutStatus (&cpu_status);
+  regY = 1;
+  ClrN;
 }
 
 void Device_EHOPEN (void)
