@@ -1,22 +1,31 @@
 #include <stdio.h>
 #include <ctype.h>
-#include <fcntl.h>
 
 #ifdef VMS
+#include <unixio.h>
 #include <file.h>
+#else
+#include <fcntl.h>
 #endif
 
-static char *rcsid = "$Id: monitor.c,v 1.2 1995/12/30 22:11:10 david Exp $";
+static char *rcsid = "$Id: monitor.c,v 1.8 1996/07/23 01:21:12 david Exp $";
 
-#include "system.h"
 #include "atari.h"
 #include "cpu.h"
+#include "antic.h"
+#include "gtia.h"
 
 #define FALSE   0
 #define TRUE    1
 
+extern int count[256];
+
 extern UBYTE	memory[65536];
 
+#ifdef TRACE
+int tron = FALSE;
+#endif
+
 unsigned int disassemble (UWORD addr1, UWORD addr2);
 void show_opcode (UBYTE instr);
 void show_operand (UBYTE instr);
@@ -70,8 +79,6 @@ int get_hex (char *string, UWORD *hexval)
 
 int monitor ()
 {
-  CPU_Status	cpu_status;
-
   UWORD	addr;
   UWORD	break_addr;
   char	s[128];
@@ -79,7 +86,7 @@ int monitor ()
 
   addr = 0;
 
-  CPU_GetStatus (&cpu_status);
+  CPU_GetStatus ();
 
   while (TRUE)
     {
@@ -100,6 +107,11 @@ int monitor ()
 	}
       else if (strcmp(t,"CONT") == 0)
 	{
+	  int i;
+
+	  for (i=0;i<256;i++)
+	    count[i] = 0;
+
 	  return 1;
 	}
       else if (strcmp(t,"DLIST") == 0)
@@ -173,78 +185,133 @@ int monitor ()
 	{
 	  get_hex (NULL, &addr);
 	  
-	  cpu_status.PC = addr;
+	  regPC = addr;
 	}
       else if (strcmp(t,"SETS") == 0)
 	{
 	  get_hex (NULL, &addr);
-	  cpu_status.S = addr & 0xff;
+	  regS = addr & 0xff;
 	}
       else if (strcmp(t,"SETA") == 0)
 	{
 	  get_hex (NULL, &addr);
-	  cpu_status.A = addr & 0xff;
+	  regA = addr & 0xff;
 	}
       else if (strcmp(t,"SETX") == 0)
 	{
 	  get_hex (NULL, &addr);
-	  cpu_status.X = addr & 0xff;
+	  regX = addr & 0xff;
 	}
       else if (strcmp(t,"SETY") == 0)
 	{
 	  get_hex (NULL, &addr);
-	  cpu_status.Y = addr & 0xff;
+	  regY = addr & 0xff;
 	}
       else if (strcmp(t,"SETN") == 0)
 	{
 	  get_hex (NULL, &addr);
-	  cpu_status.flag.N = addr;
+	  if (addr)
+	    SetN;
+	  else
+	    ClrN;
 	}
       else if (strcmp(t,"SETV") == 0)
 	{
 	  get_hex (NULL, &addr);
-	  cpu_status.flag.V = addr;
+	  if (addr)
+	    SetV;
+	  else
+	    ClrV;
 	}
       else if (strcmp(t,"SETB") == 0)
 	{
 	  get_hex (NULL, &addr);
-	  cpu_status.flag.B = addr;
+	  if (addr)
+	    SetB;
+	  else
+	    ClrB;
 	}
       else if (strcmp(t,"SETD") == 0)
 	{
 	  get_hex (NULL, &addr);
-	  cpu_status.flag.D = addr;
+	  if (addr)
+	    SetD;
+	  else
+	    ClrD;
 	}
       else if (strcmp(t,"SETI") == 0)
 	{
 	  get_hex (NULL, &addr);
-	  cpu_status.flag.I = addr;
+	  if (addr)
+	    SetI;
+	  else
+	    ClrI;
 	}
       else if (strcmp(t,"SETZ") == 0)
 	{
 	  get_hex (NULL, &addr);
-	  cpu_status.flag.Z = addr;
+	  if (addr)
+	    SetZ;
+	  else
+	    ClrZ;
 	}
       else if (strcmp(t,"SETC") == 0)
 	{
 	  get_hex (NULL, &addr);
-	  cpu_status.flag.C = addr;
+	  if (addr)
+	    SetC;
+	  else
+	    ClrC;
+	}
+#ifdef TRACE
+      else if (strcmp(t,"TRON") == 0)
+	tron = TRUE;
+      else if (strcmp(t,"TROFF") == 0)
+	tron = FALSE;
+#endif
+      else if (strcmp(t,"PROFILE") == 0)
+	{
+	  int i;
+
+	  for (i=0;i<10;i++)
+	    {
+	      int max, instr;
+	      int j;
+
+	      max = count[0];
+	      instr = 0;
+
+	      for (j=1;j<256;j++)
+		{
+		  if (count[j] > max)
+		    {
+		      max = count[j];
+		      instr = j;
+		    }
+		}
+
+	      if (max > 0)
+		{
+		  count[instr] = 0;
+		  printf ("%02x has been executed %d times\n",
+			  instr, max);
+		}
+	      else
+		{
+		  printf ("Instruction Profile data not available\n");
+		  break;
+		}
+	    }
 	}
       else if (strcmp(t,"SHOW") == 0)
 	{
-	  printf ("PC=%4x, A=%2x, S=%2x, X=%2x, Y=%2x, C=%1x, Z=%1x, I=%1x, D=%1x, B=%1x, V=%1x, N=%1x\n",
-		  cpu_status.PC,
-		  cpu_status.A,
-		  cpu_status.S,
-		  cpu_status.X,
-		  cpu_status.Y,
-		  cpu_status.flag.C,
-		  cpu_status.flag.Z,
-		  cpu_status.flag.I,
-		  cpu_status.flag.D,
-		  cpu_status.flag.B,
-		  cpu_status.flag.V,
-		  cpu_status.flag.N);
+	  printf ("PC=%04x, A=%02x, S=%02x, X=%02x, Y=%02x, P=%02x\n",
+		  regPC,
+		  regA,
+		  regS,
+		  regX,
+		  regY,
+		  regP);
 	}
       else if (strcmp(t,"ROM") == 0)
 	{
@@ -439,6 +506,34 @@ int monitor ()
 
 	  addr = disassemble (addr1,addr2);
 	}
+      else if (strcmp(t,"ANTIC") == 0)
+	{
+	  printf ("DMACTL=%02x    CHACTL=%02x    DLISTL=%02x    "
+		  "DLISTH=%02x    HSCROL=%02x    VSCROL=%02x\n",
+		  DMACTL, CHACTL, DLISTL, DLISTH, HSCROL, VSCROL);
+	  printf ("PMBASE=%02x    CHBASE=%02x    WSYNC= xx    "
+		  "VCOUNT=%02x    PENH=  xx    PENV=  xx\n",
+		  PMBASE, CHBASE, (ypos + 8) >> 1);
+	  printf ("NMIEN= %02x    NMIRES=xx\n", NMIEN);
+	}
+      else if (strcmp(t,"GTIA") == 0)
+	{
+	  printf ("HPOSP0=%02x    HPOSP1=%02x    HPOSP2=%02x    "
+		  "HPOSP3=%02x    HPOSM0=%02x    HPOSM1=%02x\n",
+		  HPOSP0, HPOSP1, HPOSP2, HPOSP3, HPOSM0, HPOSM1);
+	  printf ("HPOSM2=%02x    HPOSM3=%02x    SIZEP0=%02x    "
+		  "SIZEP1=%02x    SIZEP2=%02x    SIZEP3=%02x\n",
+		  HPOSM2, HPOSM3, SIZEP0, SIZEP1, SIZEP2, SIZEP3);
+	  printf ("SIZEM= %02x    GRAFP0=%02x    GRAFP1=%02x    "
+		  "GRAFP2=%02x    GRAFP3=%02x    GRAFM =%02x\n",
+		  SIZEM, GRAFP0, GRAFP1, GRAFP2, GRAFP3, GRAFM);
+	  printf ("COLPM0=%02x    COLPM1=%02x    COLPM2=%02x    "
+		  "COLPM3=%02x    COLPF0=%02x    COLPF1=%02x\n",
+		  COLPM0, COLPM1, COLPM2, COLPM3, COLPF0, COLPF1);
+	  printf ("COLPF2=%02x    COLPF3=%02x    COLBK= %02x    "
+		  "PRIOR= %02x    GRACTL=%02x\n",
+		  COLPF2, COLPF3, COLBK, PRIOR, GRACTL);
+	}
       else if (strcmp(t,"HELP") == 0 || strcmp(t,"?") == 0)
 	{
 	  printf ("SET{PC,A,S,X,Y} hexval    - Set Register Value\n");
@@ -452,6 +547,12 @@ int monitor ()
 	  printf ("SHOW                      - Show Registers\n");
 	  printf ("WRITE startaddr endaddr   - Write specified area of memory to memdump.dat\n");
 	  printf ("SUM [startaddr] [endaddr] - SUM of specified memory range\n");
+#ifdef TRACE
+	  printf ("TRON                      - Trace On\n");
+	  printf ("TROFF                     - Trace Off\n");
+#endif
+          printf ("ANTIC                     - Display ANTIC registers\n");
+          printf ("GTIA                      - Display GTIA registers\n");
 	  printf ("QUIT                      - Quit Emulation\n");
 	  printf ("HELP or ?                 - This Text\n");
 	}
